﻿Imports System.Text

''' <summary>
''' This is the calculator "engine"... it processes keystrokes and
''' performs the appropriate functions
''' </summary>
''' <remarks></remarks>
Public Class Calculator

    Private Const F_KEY As Integer = 128
    Private Const G_KEY As Integer = 256

    ''' <summary>
    ''' The native key codes
    ''' </summary>
    ''' <remarks>
    ''' Digits (0-9 and A-F) use their own values, all others use a grid system like this:
    '''  A,  B,  C,  D,  E,  F, 7,  8,  9, 10
    ''' 21, 22, 23, 24, 25, 26, 4,  5,  6, 20
    ''' 31, 32, 33, 35, 35, 36, 1,  2,  3, 30
    ''' 41, 42, 43, 44, 45,   , 0, 48, 49, 40
    ''' For simplicity's sake, I just treat key codes as hex
    '''</remarks>
    Public Enum k As Integer
        KeyA = &HA
        KeyB = &HB
        KeyC = &HC
        KeyD = &HD
        KeyE = &HE
        KeyF = &HF
        Key7 = &H7
        Key8 = &H8
        Key9 = &H9
        KeyDiv = &H10
        KeyGSB = &H21
        KeyGTO = &H22
        KeyHEX = &H23
        KeyDEC = &H24
        KeyOCT = &H25
        KeyBIN = &H26
        Key4 = &H4
        Key5 = &H5
        Key6 = &H6
        KeyMul = &H20
        KeyRS = &H31
        KeySST = &H32
        KeyRol = &H33
        KeyXY = &H34
        KeyBSP = &H35
        KeyEnt = &H36
        Key1 = &H1
        Key2 = &H2
        Key3 = &H3
        KeyMin = &H30
        KeyON = &H41
        KeyFKey = &H42
        KeyGKey = &H43
        KeySTO = &H44
        KeyRCL = &H45
        Key0 = &H0
        KeyDp = &H48
        KeyCHS = &H49
        KeyAdd = &H40

        ' The Yellow (F key) shifted functions
        FnSL = k.KeyA + F_KEY
        FnSR = k.KeyB + F_KEY
        FnRL = k.KeyC + F_KEY
        FnRR = k.KeyD + F_KEY
        FnRLn = k.KeyE + F_KEY
        FnRRn = k.KeyF + F_KEY
        FnMASKL = k.Key7 + F_KEY
        FnMASKR = k.Key8 + F_KEY
        FnRMD = k.Key9 + F_KEY
        FnXOR = k.KeyDiv + F_KEY
        FnXIndex = k.KeyGSB + F_KEY
        FnXI = k.KeyGTO + F_KEY
        FnShowHex = k.KeyHEX + F_KEY
        FnShowDec = k.KeyDEC + F_KEY
        FnShowOct = k.KeyOCT + F_KEY
        FnShowBin = k.KeyBIN + F_KEY
        FnSB = k.Key4 + F_KEY
        FnCB = k.Key5 + F_KEY
        FnBSet = k.Key6 + F_KEY
        FnAND = k.KeyMul + F_KEY
        FnIndex = k.KeyRS + F_KEY
        FnI = k.KeySST + F_KEY
        FnClearPrgm = k.KeyRol + F_KEY
        FnClearReg = k.KeyXY + F_KEY
        FnClearPrefix = k.KeyBSP + F_KEY
        FnWINDOW = k.KeyEnt + F_KEY
        FnSet1s = k.Key1 + F_KEY
        FnSet2s = k.Key2 + F_KEY
        FnSetUnsgn = k.Key3 + F_KEY
        FnNOT = k.KeyMin + F_KEY
        FnWSIZE = k.KeySTO + F_KEY
        FnFLOAT = k.KeyRCL + F_KEY
        FnMEM = k.Key0 + F_KEY
        FnSTATUS = k.KeyDp + F_KEY
        FnEEX = k.KeyCHS + F_KEY
        FnOR = k.KeyAdd + F_KEY

        ' The Blue (G key) shifted functions
        FnLJ = k.KeyA + G_KEY
        FnASR = k.KeyB + G_KEY
        FnRLC = k.KeyC + G_KEY
        FnRRC = k.KeyD + G_KEY
        FnRLCn = k.KeyE + G_KEY
        FnRRCn = k.KeyF + G_KEY
        FnNumB = k.Key7 + G_KEY
        FnABS = k.Key8 + G_KEY
        FnDBLR = k.Key9 + G_KEY
        FnDBLDiv = k.KeyDiv + G_KEY
        FnRTN = k.KeyGSB + G_KEY
        FnLBL = k.KeyGTO + G_KEY
        FnDSZ = k.KeyHEX + G_KEY
        FnISZ = k.KeyDEC + G_KEY
        FnSqrt = k.KeyOCT + G_KEY
        FnInv = k.KeyBIN + G_KEY
        FnSF = k.Key4 + G_KEY
        FnCF = k.Key5 + G_KEY
        FnFSet = k.Key6 + G_KEY
        FnDBLMul = k.KeyMul + G_KEY
        FnPR = k.KeyRS + G_KEY
        FnBST = k.KeySST + G_KEY
        FnRolUp = k.KeyRol + G_KEY
        FnPSE = k.KeyXY + G_KEY
        FnCLX = k.KeyBSP + G_KEY
        FnLSTX = k.KeyEnt + G_KEY
        FnXlteY = k.Key1 + G_KEY
        FnXlt0 = k.Key2 + G_KEY
        FnXgtY = k.Key3 + G_KEY
        FnXgt0 = k.KeyMin + G_KEY
        FnSLeft = k.KeySTO + G_KEY
        FnSRight = k.KeyRCL + G_KEY
        FnXneY = k.Key0 + G_KEY
        FnXne0 = k.KeyDp + G_KEY
        FnXeqY = k.KeyCHS + G_KEY
        FnXeq0 = k.KeyAdd + G_KEY
    End Enum

    Private cs As CalcState
    Private Packet As DisplayPacket
    Private RawDisplay As StringBuilder
    Private WinPos As Integer

    Private ShortCutAllowed() As k = {k.KeySTO, k.KeyRCL, k.KeyGSB, k.KeyGTO}
    Private DpNotAllowed() As k = {k.FnWINDOW, k.FnSF, k.FnCF, k.FnFSet, k.FnLBL, k.KeyGSB}
    Private TerminateInput() As k = {k.KeyBSP, k.KeyDp, k.KeyCHS, k.FnEEX}
    Private DoNotResetWinPos() As k = {k.FnSLeft, k.FnSRight, k.KeySTO, k.FnMEM, k.FnSTATUS, k.FnPR, k.FnLBL, k.FnRTN, k.FnSF,
                                      k.FnCF, k.FnFSet, k.FnPSE, k.KeyRS, k.KeyGSB, k.KeyGTO, k.KeySST, k.FnABS, k.FnDSZ, k.FnISZ,
                                      k.FnShowBin, k.FnClearPrgm, k.FnClearReg, k.FnClearPrefix, k.FnXeq0, k.FnXeqY, k.FnXgt0,
                                      k.FnXgtY, k.FnXlt0, k.FnXlteY, k.FnXne0, k.FnXneY, k.FnWINDOW}

    Private Prefix As Integer
    Private Stage As Stack(Of Integer)
    Private PrgmEntry As Boolean
    Private StackDisable As Boolean
    Private PadDecimal As Boolean
    Private previousRaw As String

    Private numRegs As Integer
    Private syncConv As Boolean
    Private numLines As Integer
    Private sleepDelay As Integer

    ''' <summary>
    ''' The default constructor
    ''' </summary>
    ''' <param name="calc_state">The calculate "state" object</param>
    ''' <remarks>
    ''' This class changes the text in the tbDisplay control on the
    ''' main form.
    ''' </remarks>
    Public Sub New(calc_state As CalcState)
        cs = calc_state
        RawDisplay = New StringBuilder
        WinPos = 0

        Prefix = 0
        Stage = New Stack(Of Integer)
        PrgmEntry = False
        StackDisable = False
        PadDecimal = False

        Packet = New DisplayPacket
        Packet.Carry_Annunciator = cs.Flags(CalcState.CalcFlag.Carry)
        Packet.Overflow_Annunciator = cs.Flags(CalcState.CalcFlag.Overflow)
        Packet.DisplayText = FormatDisplay()

        ' cache the property values
        numRegs = My.Settings.NumRegisters
        sleepDelay = My.Settings.SleepDelay
        numLines = My.Settings.PrgmMemoryLines
        syncConv = My.Settings.SyncConversions

    End Sub

    ''' <summary>
    ''' Process a calculator key stroke
    ''' </summary>
    ''' <param name="CalcKey">A key value</param>\
    ''' <returns>A "display packet" filled with information for the GUI</returns>
    ''' <remarks>
    ''' Accepts both key values and function values (those with
    ''' prefix already applied)
    ''' </remarks>
    Public Function ProcessKey(CalcKey As Integer) As DisplayPacket
        Dim key As k
        Dim val As Integer

        ' There are 3 types of "keys" that we deal with here:
        ' 1) Ordinary "white" keys (like the plus key for addition)
        ' 2) Shifted functions (the yellow or blue functions) that
        '    require the yellow f or blue g key followed by an ordinary key
        ' 3) Multi-stage functions that require 1 or more additional
        '    keystrokes as operands.  Example: the STO function
        '    can't be processed until a 2nd key is pressed to indicate
        '    the register number.  Note: Multi-stage can also be
        '    shifted.

        ' Shifted functions are handled by adding the prefix value
        If CalcKey = k.KeyFKey Or CalcKey = k.KeyGKey Then
            ' in case we accidently "double" shifted
            key = CType(CalcKey, k)
        Else
            key = CType(CalcKey + Prefix, k)
        End If

        ' clean up the packet for this next run
        Packet.AlternateText = String.Empty
        Packet.F_Annunciator = False
        Packet.G_Annunciator = False
        Packet.Prgm_Annunciator = PrgmEntry
        Packet.MenuNeedsUpdating = False
        Packet.Delay = 0
        Packet.Beep = False
        Packet.Start = DisplayPacket.StartType.None

        ' This is the main loop for processing the keys and functions
        Select Case key
            Case k.KeyA               ' The A key (for hex numbers)
                ' Is this just a part of a multi-stage command?
                If Stage.Count > 0 Then
                    Stage.Push(k.KeyA)

                    ' return to the first stage for processing
                    Return ProcessKey(Stage.ToArray()(Stage.Count - 1))
                End If

                ' Is this valid for the current mode?
                If cs.OpMode <> CalcState.CalcOpMode.Hex Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine("A")
                Else
                    RawDisplay.Append("A")
                End If
            Case k.KeyB               ' The B key (for hex numbers)
                ' Is this just a part of a multi-stage command?
                If Stage.Count > 0 Then
                    Stage.Push(k.KeyB)

                    ' return to the first stage for processing
                    Return ProcessKey(Stage.ToArray()(Stage.Count - 1))
                End If

                ' Is this valid for the current mode?
                If cs.OpMode <> CalcState.CalcOpMode.Hex Then
                    Packet.Beep = True
                    Exit Select
                End If
                If PrgmEntry Then
                    recordLine("B")
                Else
                    RawDisplay.Append("B")
                End If
            Case k.KeyC               ' The C key (for hex numbers)
                ' Is this just a part of a multi-stage command?
                If Stage.Count > 0 Then
                    Stage.Push(k.KeyC)

                    ' return to the first stage for processing
                    Return ProcessKey(Stage.ToArray()(Stage.Count - 1))
                End If

                ' Is this valid for the current mode?
                If cs.OpMode <> CalcState.CalcOpMode.Hex Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine("C")
                Else
                    RawDisplay.Append("C")
                End If
            Case k.KeyD               ' The D key (for hex numbers)
                ' Is this just a part of a multi-stage command?
                If Stage.Count > 0 Then
                    Stage.Push(k.KeyD)

                    ' return to the first stage for processing
                    Return ProcessKey(Stage.ToArray()(Stage.Count - 1))
                End If

                ' Is this valid for the current mode?
                If cs.OpMode <> CalcState.CalcOpMode.Hex Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine("D")
                Else
                    RawDisplay.Append("D")
                End If
            Case k.KeyE               ' The E key (for hex numbers)
                ' Is this just a part of a multi-stage command?
                If Stage.Count > 0 Then
                    Stage.Push(k.KeyE)

                    ' return to the first stage for processing
                    Return ProcessKey(Stage.ToArray()(Stage.Count - 1))
                End If

                ' Is this valid for the current mode?
                If cs.OpMode <> CalcState.CalcOpMode.Hex Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine("E")
                Else
                    RawDisplay.Append("E")
                End If
            Case k.KeyF               ' The F key (for hex numbers)
                ' Is this just a part of a multi-stage command?
                If Stage.Count > 0 Then
                    Stage.Push(k.KeyF)

                    ' return to the first stage for processing
                    Return ProcessKey(Stage.ToArray()(Stage.Count - 1))
                End If

                ' Is this valid for the current mode?
                If cs.OpMode <> CalcState.CalcOpMode.Hex Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine("F")
                Else
                    RawDisplay.Append("F")
                End If
            Case k.Key7               ' The 7 key
                ' Is this just a part of a multi-stage command?
                If Stage.Count > 0 Then
                    Stage.Push(k.Key7)

                    ' return to the first stage for processing
                    Return ProcessKey(Stage.ToArray()(Stage.Count - 1))
                End If

                ' Is this valid for the current mode?
                If cs.OpMode > CalcState.CalcOpMode.Oct Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine("7")
                Else
                    RawDisplay.Append("7")
                End If
            Case k.Key8               ' The 8 key
                ' Is this just a part of a multi-stage command?
                If Stage.Count > 0 Then
                    Stage.Push(k.Key8)

                    ' return to the first stage for processing
                    Return ProcessKey(Stage.ToArray()(Stage.Count - 1))
                End If

                ' Is this valid for the current mode?
                If cs.OpMode > CalcState.CalcOpMode.Dec Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine("8")
                Else
                    RawDisplay.Append("8")
                End If
            Case k.Key9               ' The 9 key
                ' Is this just a part of a multi-stage command?
                If Stage.Count > 0 Then
                    Stage.Push(k.Key9)

                    ' return to the first stage for processing
                    Return ProcessKey(Stage.ToArray()(Stage.Count - 1))
                End If

                ' Is this valid for the current mode?
                If cs.OpMode > CalcState.CalcOpMode.Dec Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine("9")
                Else
                    RawDisplay.Append("9")
                End If
            Case k.KeyDiv             ' Division
                If PrgmEntry Then
                    recordLine(CInt(key), "/")
                Else
                    Dim temp As New Register(cs.WordSize, cs.ArithMode)
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        ' v6.0.7 - 22 Jul 2015
                        If cs.Stack.X.FVal = 0 Then
                            Packet.AlternateText = "Error 0 - Improper Math Operation"
                            Return Packet
                        End If

                        Dim y As Double
                        y = cs.Stack.Pop.FVal
                        temp.FVal = cs.Stack.Pop.FVal / y
                        cs.Flags(CalcState.CalcFlag.Overflow) = Double.IsInfinity(temp.FVal)
                    Else
                        ' v6.0.7 - 22 Jul 2015
                        If BigInt.AbsoluteValue(cs.Stack.X.BiVal).IsZero Then
                            Packet.AlternateText = "Error 0 - Improper Math Operation"
                            Return Packet
                        End If

                        Dim y As BigInt
                        y = cs.Stack.Pop.BiVal
                        temp.BiVal = BigInt.Divide(cs.Stack.Pop.BiVal, y)
                        cs.Flags(CalcState.CalcFlag.Carry) = temp.BiVal.CarryBit
                        cs.Flags(CalcState.CalcFlag.Overflow) = temp.BiVal.Overflow
                    End If
                    Packet.MenuNeedsUpdating = True
                    cs.Stack.Push(temp)
                End If
            Case k.KeyGSB             ' Go to Subroutine (Program Mode)
                ' This is a multi-stage command, so we have to store 1
                ' more keystroke before we can do anything
                If Stage.Count = 0 Then
                    Stage.Push(k.KeyGSB)
                    Exit Select
                End If

                ' Get the stage operand
                val = Stage.Pop
                Stage.Clear()

                If PrgmEntry Then
                    recordLine(CInt(key), val, "GSB " & Hex(val))
                Else
                    StackDisable = False

                    ' A quick sanity check
                    If cs.PrgmMemory.Count = 0 Then
                        Packet.AlternateText = "Error 4 - Improper Label"
                        Return Packet
                    End If

                    ' find the starting location
                    Dim start As Integer = FindLabel(cs.PrgmPosition, val)
                    If start < 0 Then
                        Packet.AlternateText = "Error 4 - Improper Label"
                        Return Packet
                    End If

                    ' Store the current location for an eventual RTN
                    ' Note: Unlike the real calculator, I did not limit the stack depth
                    ' to 4 (so there is no "Error 5 - Subroutine Level Too Deep")
                    cs.PrgmRetStack.Push(cs.PrgmPosition)
                    cs.PrgmPosition = start

                    ' start a program
                    If Not cs.PrgmRunning Then
                        cs.PrgmRunning = True
                        cs.PrgmRetStack.Clear()

                        Packet.AlternateText = "Running"
                        Packet.Start = DisplayPacket.StartType.RunProgram
                    End If
                End If
            Case k.KeyGTO             ' Go To Label (Program/Record Mode)
                ' There are two version of GTO... during execution of a program
                ' it accepts a single operand.  However, during editing of a program
                ' GTO accepts a decimal point and 3 digits (for the line number)
                If Stage.Count = 0 Then
                    Stage.Push(k.KeyGTO)
                    Exit Select
                End If

                ' check for the decimal point as the 2nd item
                If Stage.Count >= 2 And Stage.ToArray()(Stage.Count - 2) = k.KeyDp Then
                    If Stage.Count < 5 Then
                        ' we're not done yet... we need more digits
                        Exit Select
                    Else
                        ' decode the 3-digit line number
                        Dim digit1, digit2, digit3, line As Integer
                        digit3 = Stage.Pop
                        digit2 = Stage.Pop
                        digit1 = Stage.Pop

                        Stage.Clear()
                        ' do some sanity checking for digit entry
                        If digit1 > 9 Or digit2 > 9 Or digit3 > 9 Then
                            Packet.AlternateText = "Error 1 - Improper GTO. Number"
                            Return Packet
                        End If
                        line = (digit1 * 100) + (digit2 * 10) + digit3

                        ' and some more checking on the decoded number
                        If line < 0 Or line > cs.PrgmMemory.Count Then
                            Packet.AlternateText = "Error 4 - Improper Line Number"
                            Return Packet
                        End If

                        cs.PrgmPosition = line
                        Exit Select
                    End If
                End If

                ' Get the stage operand
                val = Stage.Pop
                Stage.Clear()

                If PrgmEntry Then
                    recordLine(CInt(key), val, "GTO " & Hex(val))
                Else
                    val = FindLabel(cs.PrgmPosition, val)
                    If val < 0 Then
                        cs.PrgmRunning = False
                        Prefix = 0
                        Packet.AlternateText = "Error 4 - Improper Label"
                        Return Packet
                    End If
                    cs.PrgmPosition = val
                End If
                ' Note: the "GTO.nnn" version can not be stored in Program Mode
            Case k.KeyHEX             ' Hexidecimal Mode
                If PrgmEntry Then
                    recordLine(CInt(key), "HEX")
                End If

                ' This is a bit strange... if you are entering a program, then
                ' you also need to convert to that mode.
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    cs.SyncValues()
                End If
                cs.OpMode = CalcState.CalcOpMode.Hex
                Packet.MenuNeedsUpdating = True
            Case k.KeyDEC             ' Decimal Mode
                If PrgmEntry Then
                    recordLine(CInt(key), "DEC")
                End If

                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    cs.SyncValues()
                End If
                cs.OpMode = CalcState.CalcOpMode.Dec
                Packet.MenuNeedsUpdating = True
            Case k.KeyOCT             ' Octal Mode
                If PrgmEntry Then
                    recordLine(CInt(key), "OCT")
                End If

                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    cs.SyncValues()
                End If
                cs.OpMode = CalcState.CalcOpMode.Oct
                Packet.MenuNeedsUpdating = True
            Case k.KeyBIN             ' Binary Mode
                If PrgmEntry Then
                    recordLine(CInt(key), "BIN")
                End If

                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    cs.SyncValues()
                End If
                cs.OpMode = CalcState.CalcOpMode.Bin
                Packet.MenuNeedsUpdating = True
            Case k.Key4               ' The 4 key
                ' Is this just a part of a multi-stage command?
                If Stage.Count > 0 Then
                    Stage.Push(k.Key4)

                    ' return to the first stage for processing
                    Return ProcessKey(Stage.ToArray()(Stage.Count - 1))
                End If

                ' Is this valid for the current mode?
                If cs.OpMode > CalcState.CalcOpMode.Oct Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine("4")
                Else
                    RawDisplay.Append("4")
                End If
            Case k.Key5               ' The 5 key
                ' Is this just a part of a multi-stage command?
                If Stage.Count > 0 Then
                    Stage.Push(k.Key5)

                    ' return to the first stage for processing
                    Return ProcessKey(Stage.ToArray()(Stage.Count - 1))
                End If

                ' Is this valid for the current mode?
                If cs.OpMode > CalcState.CalcOpMode.Oct Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine("5")
                Else
                    RawDisplay.Append("5")
                End If
            Case k.Key6               ' The 6 key
                ' Is this just a part of a multi-stage command?
                If Stage.Count > 0 Then
                    Stage.Push(k.Key6)

                    ' return to the first stage for processing
                    Return ProcessKey(Stage.ToArray()(Stage.Count - 1))
                End If

                ' Is this valid for the current mode?
                If cs.OpMode > CalcState.CalcOpMode.Oct Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine("6")
                Else
                    RawDisplay.Append("6")
                End If
            Case k.KeyMul             ' Multiplication
                If PrgmEntry Then
                    recordLine(CInt(key), "*")
                Else
                    Dim temp As New Register(cs.WordSize, cs.ArithMode)
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        temp.FVal = cs.Stack.Pop.FVal * cs.Stack.Pop.FVal
                        cs.Flags(CalcState.CalcFlag.Overflow) = Double.IsInfinity(temp.FVal)
                    Else
                        temp.BiVal = BigInt.Multiply(cs.Stack.Pop.BiVal, cs.Stack.Pop.BiVal)
                        cs.Flags(CalcState.CalcFlag.Overflow) = temp.BiVal.Overflow
                    End If
                    Packet.MenuNeedsUpdating = True
                    cs.Stack.Push(temp)
                End If
            Case k.KeyRS              ' Run / Stop (Program Mode)
                ' There is a strange shortcut allowed here...  When used as an operand
                ' with a function, then you can ignore the Yellow shift (f) key
                ' and still get the affect of k.FnIndex (rather than k.KeyRS)
                ' v7.0.1 - 3 Nov 21: bug in casting the type
                If Stage.Count = 1 AndAlso Array.IndexOf(ShortCutAllowed, CType(Stage.Peek, k)) >= 0 Then
                    Stage.Push(k.FnIndex)

                    ' return to the first stage for processing
                    Return ProcessKey(Stage.ToArray()(Stage.Count - 1))
                End If

                If PrgmEntry Then
                    recordLine(CInt(key), "R/S")
                Else
                    ' a quick sanity check
                    If cs.PrgmMemory.Count > 0 Then
                        If cs.PrgmRunning Then
                            ' if inside a program, the only option is to stop
                            cs.PrgmRunning = False
                        Else
                            ' If outside a program, the only option is to start
                            ' since any keystroke would have stopped a running
                            ' application.
                            cs.PrgmRunning = True
                            Packet.AlternateText = "Running"
                            Packet.Start = DisplayPacket.StartType.RunProgram
                        End If
                    End If
                End If
            Case k.KeySST             ' Single step (Program Mode)
                ' There is a strange shortcut allowed here...  When used as an operand
                ' with a function, then you can ignore the Yellow shift (f) key
                ' and still get the affect of k.FnI (rather than k.KeySST)
                ' v7.0.1 - 3 Nov 21: bug in casting the type
                If Stage.Count = 1 AndAlso Array.IndexOf(ShortCutAllowed, CType(Stage.Peek, k)) >= 0 Then
                    Stage.Push(k.FnI)

                    ' return to the first stage for processing
                    Return ProcessKey(Stage.ToArray()(Stage.Count - 1))
                End If

                ' In PrgmEntry mode, we just increment the position, but in the
                ' run mode, we actually execute the instructions
                If PrgmEntry Then
                    cs.PrgmPosition += 1
                    If cs.PrgmPosition > cs.PrgmMemory.Count Then
                        cs.PrgmPosition = 0
                    End If
                Else
                    ' a quick sanity check
                    If cs.PrgmMemory.Count > 0 Then
                        ' v6.0.2 - 26 Apr 12
                        If cs.PrgmPosition >= cs.PrgmMemory.Count Then
                            cs.PrgmPosition = 0
                        End If

                        ' temporarily show the program codes
                        Packet.AlternateText = (cs.PrgmPosition + 1).ToString("000") & "- " & Left(cs.PrgmMemory(cs.PrgmPosition), 8)
                        Packet.Delay = CInt(sleepDelay / 2)

                        ' run one line
                        Packet.Start = DisplayPacket.StartType.RunLine

                        cs.PrgmPosition += 1
                        If cs.PrgmPosition >= cs.PrgmMemory.Count Then
                            cs.PrgmPosition = 0
                        End If
                    End If
                End If
                ' Note: SST can not be stored in Program Mode
            Case k.KeyRol             ' Roll the stack
                If PrgmEntry Then
                    recordLine(CInt(key), "Rv")
                Else
                    Dim temp As Register
                    temp = cs.Stack.X.Copy
                    cs.Stack.X = cs.Stack.Y.Copy
                    cs.Stack.Y = cs.Stack.Z.Copy
                    cs.Stack.Z = cs.Stack.T.Copy
                    cs.Stack.T = temp
                    StackDisable = False
                End If
            Case k.KeyXY              ' Exchange X and Y
                If PrgmEntry Then
                    recordLine(CInt(key), "X:Y")
                Else
                    Dim temp As Register
                    temp = cs.Stack.Y.Copy
                    cs.Stack.Y = cs.Stack.X.Copy
                    cs.Stack.X = temp
                    StackDisable = False
                End If
            Case k.KeyBSP               ' The Backspace Key
                If PrgmEntry Then
                    If cs.PrgmPosition > 0 Then
                        cs.PrgmMemory.RemoveAt(cs.PrgmPosition - 1)
                        cs.PrgmPosition = cs.PrgmPosition - 1
                    End If
                Else
                    ' If digit entry has already been terminated, then this
                    ' acts just like CLx
                    If RawDisplay.Length = 0 Then
                        cs.Stack.X = New Register(cs.WordSize, cs.ArithMode)
                        StackDisable = True
                    Else
                        RawDisplay.Remove(RawDisplay.Length - 1, 1)
                        If RawDisplay.Length = 0 Then
                            cs.Stack.X = New Register(cs.WordSize, cs.ArithMode)
                        End If

                        ' Since we allow you to backspace over a number
                        ' in scientific notation, we must make sure the
                        ' raw display will still parse
                        Dim temp As String = RawDisplay.ToString
                        If temp.EndsWith("e") Then
                            RawDisplay.Remove(RawDisplay.Length - 1, 1)
                        ElseIf temp.EndsWith("e-") Then
                            RawDisplay.Remove(RawDisplay.Length - 2, 2)
                        End If

                        ' What about leaving a "naked" minus sign
                        If temp = "-" Then
                            RawDisplay.Length = 0
                            cs.Stack.X = New Register(cs.WordSize, cs.ArithMode)
                        End If

                        ' Here is a strange twist... if we only have 1 digit remaining
                        ' then we have to explicitly tell the code that follows
                        ' this select statement that this 1 digit is not a new entry.
                        ' We do with via the StackDisable variable (although this has
                        ' nothing to do with disabling the stack).
                        If RawDisplay.Length = 1 Then
                            StackDisable = True
                        End If
                    End If
                End If
                ' Note: BSP can not be stored in Program Mode
            Case k.KeyEnt             ' The Enter key
                If PrgmEntry Then
                    recordLine(CInt(key), "Enter")
                Else
                    ' push the X value on the stack
                    cs.Stack.Push(cs.Stack.X.Copy)
                    ' "disable" the stack for the next entry
                    StackDisable = True
                End If
            Case k.Key1               ' The 1 key
                ' Is this just a part of a multi-stage command?
                If Stage.Count > 0 Then
                    Stage.Push(k.Key1)

                    ' return to the first stage for processing
                    Return ProcessKey(Stage.ToArray()(Stage.Count - 1))
                End If

                If PrgmEntry Then
                    recordLine("1")
                Else
                    RawDisplay.Append("1")
                End If
            Case k.Key2               ' The 2 key
                ' Is this just a part of a multi-stage command?
                If Stage.Count > 0 Then
                    Stage.Push(k.Key2)

                    ' return to the first stage for processing
                    Return ProcessKey(Stage.ToArray()(Stage.Count - 1))
                End If

                ' Is this valid for the current mode?
                If cs.OpMode > CalcState.CalcOpMode.Oct Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine("2")
                Else
                    RawDisplay.Append("2")
                End If
            Case k.Key3               ' The 3 key
                ' Is this just a part of a multi-stage command?
                If Stage.Count > 0 Then
                    Stage.Push(k.Key3)

                    ' return to the first stage for processing
                    Return ProcessKey(Stage.ToArray()(Stage.Count - 1))
                End If

                ' Is this valid for the current mode?
                If cs.OpMode > CalcState.CalcOpMode.Oct Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine("3")
                Else
                    RawDisplay.Append("3")
                End If
            Case k.KeyMin             ' Subtraction
                If PrgmEntry Then
                    recordLine(CInt(key), "-")
                Else
                    Dim temp As New Register(cs.WordSize, cs.ArithMode)
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        Dim y As Double = cs.Stack.Pop.FVal
                        temp.FVal = cs.Stack.Pop.FVal - y
                        cs.Flags(CalcState.CalcFlag.Overflow) = Double.IsInfinity(temp.FVal)
                    Else
                        Dim y As BigInt = cs.Stack.Pop.BiVal
                        temp.BiVal = BigInt.Subtract(cs.Stack.Pop.BiVal, y)
                        cs.Flags(CalcState.CalcFlag.Carry) = temp.BiVal.CarryBit
                        cs.Flags(CalcState.CalcFlag.Overflow) = temp.BiVal.Overflow
                    End If
                    Packet.MenuNeedsUpdating = True
                    cs.Stack.Push(temp)
                End If
            Case k.KeyON              ' The ON key
#If DEBUG Then
                ' used for calibrating the automatic screen widths
                Packet.DisplayText = "00000000 00000000 00000000 00000000 .b."
                Packet.F_Annunciator = True
                Packet.G_Annunciator = True
                Packet.Carry_Annunciator = True
                Packet.Overflow_Annunciator = True
                Packet.Prgm_Annunciator = True
                Return Packet
#End If
            Case k.KeyFKey            ' The yellow F prefix key
                Prefix = 128

                Packet.F_Annunciator = True
                Return Packet
            Case k.KeyGKey            ' The blue G prefix key
                Prefix = 256

                Packet.G_Annunciator = True
                Return Packet
            Case k.KeySTO             ' Store a value to a register
                ' This is a multi-stage command, so we have to store up 2 or 3
                ' more keystrokes before we can do anything
                If Stage.Count = 0 Then
                    Stage.Push(k.KeySTO)
                    Exit Select
                End If

                ' a quick sanity check
                If Stage.ToArray()(Stage.Count - 1) <> k.KeySTO Then
                    Stage.Clear()
                    Stage.Push(k.KeySTO)
                    Exit Select
                End If

                ' This is an odd, one... there could be 2 or 3 stages
                If Stage.Peek = k.KeyDp Then
                    ' we need another digit to continue
                    Exit Select
                End If

                ' convert the values in the stages into a register number
                val = Stage.Pop
                If val > (numRegs - 1) And val <> k.FnI And val <> k.FnIndex Then
                    Packet.AlternateText = "Error 3 - Improper Register Number"
                    Stage.Clear()
                    Return Packet
                End If

                If Stage.Pop = k.KeyDp Then
                    val = val + 16
                End If

                ' clear the stage, we've completed a sequence
                Stage.Clear()

                If PrgmEntry Then
                    If val = k.FnI Then
                        recordLine(CInt(key), CInt(k.KeySST), "STO I")
                    ElseIf val = k.FnIndex Then
                        recordLine(CInt(key), CInt(k.KeyRS), "STO (i)")
                    ElseIf val >= 16 Then
                        ' odd ball format not covered by recordline
                        cs.PrgmMemory.Insert(cs.PrgmPosition, String.Format("{0,5:X} .{1,1:X1}    '{2}", CInt(key), val - 16, "STO ." & Hex(val - 16)))
                        cs.PrgmPosition += 1
                    Else
                        recordLine(CInt(key), val, "STO " & Hex(val))
                    End If

                Else
                    StackDisable = False
                    ' does not effect stack
                    If val = k.FnI Then
                        ' Store into the I register
                        cs.RegIndex = cs.Stack.X.Copy
                        cs.RegIndex.BiVal.WordSize = 64
                    ElseIf val = k.FnIndex Then
                        ' Store into the register who's number is stored in I
                        Dim i As Integer
                        If cs.OpMode = CalcState.CalcOpMode.Float Then
                            i = CInt(Math.Abs(cs.RegIndex.FVal))
                        Else
                            i = BigInt.AbsoluteValue(cs.RegIndex.BiVal).ToInteger
                        End If

                        If i > numRegs - 1 Then
                            Packet.AlternateText = "Error 3 - Improper Register Number"
                            Return Packet
                        End If
                        cs.Reg(i) = cs.Stack.X.Copy
                    Else
                        ' Just store into a normal "named" register
                        cs.Reg(val) = cs.Stack.X.Copy
                    End If
                End If
            Case k.KeyRCL             ' Recall a value from a register
                ' This is a multi-stage command, so we have to store up 2 or 3
                ' more keystrokes before we can do anything
                If Stage.Count = 0 Then
                    Stage.Push(k.KeyRCL)
                    Exit Select
                End If

                ' a quick sanity check
                If Stage.ToArray()(Stage.Count - 1) <> k.KeyRCL Then
                    Stage.Clear()
                    Stage.Push(k.KeyRCL)
                    Exit Select
                End If

                ' This is an odd, one... there could be 2 or 3 stages
                If Stage.Peek = k.KeyDp Then
                    ' we need another digit to continue
                    Exit Select
                End If

                ' convert the values in the stages into a register number
                val = Stage.Pop

                If val > (numRegs - 1) And val <> k.FnI And val <> k.FnIndex Then
                    Packet.AlternateText = "Error 3 - Improper Register Number"
                    Stage.Clear()
                    Return Packet
                End If

                If Stage.Pop = k.KeyDp Then
                    val = val + 16
                End If

                Stage.Clear()

                If PrgmEntry Then
                    If val = k.FnI Then
                        recordLine(CInt(key), CInt(k.KeySST), "RCL I")
                    ElseIf val = k.FnIndex Then
                        recordLine(CInt(key), CInt(k.KeyRS), "RCL (i)")
                    ElseIf val >= 16 Then
                        ' odd ball format not covered by recordline
                        cs.PrgmMemory.Insert(cs.PrgmPosition, String.Format("{0,5:X} .{1,1:X1}    '{2}", CInt(key), val - 16, "RCL ." & Hex(val - 16)))
                        cs.PrgmPosition += 1
                    Else
                        recordLine(CInt(key), val, "RCL " & Hex(val))
                    End If
                Else
                    Dim temp As Register

                    If val = k.FnI Then
                        ' recall value from the I register
                        temp = cs.RegIndex.Copy
                    ElseIf val = k.FnIndex Then
                        ' recall the value from the register who's number is stored in I
                        Dim i As Integer
                        If cs.OpMode = CalcState.CalcOpMode.Float Then
                            i = CInt(Math.Abs(cs.RegIndex.FVal))
                        Else
                            i = BigInt.AbsoluteValue(cs.RegIndex.BiVal).ToInteger
                        End If

                        If i > numRegs - 1 Then
                            Packet.AlternateText = "Error 3 - Improper Register Number"
                            Return Packet
                        End If
                        temp = cs.Reg(i).Copy
                    Else
                        ' Just recall from a normal "named" register
                        temp = cs.Reg(val).Copy
                    End If

                    ' v6.0.6 - 1 Mar 14
                    If StackDisable Then
                        cs.Stack.X = temp
                        StackDisable = False
                    Else
                        cs.Stack.Push(temp)
                    End If
                End If
            Case k.Key0               ' The 0 zero key
                ' Is this just a part of a multi-stage command?
                If Stage.Count > 0 Then
                    Stage.Push(k.Key0)

                    ' return to the first stage for processing
                    Return ProcessKey(Stage.ToArray()(Stage.Count - 1))
                End If

                If PrgmEntry Then
                    recordLine("0")
                Else
                    RawDisplay.Append("0")
                End If
            Case k.KeyDp              ' Decimal point
                ' Is this just a part of a multi-stage command?
                If Stage.Count > 0 Then
                    ' Check for valid stages that accept a decimal point.  It makes
                    ' since to check here so we can get immediate feedback
                    ' v7.0.1 - 3 Nov 21: bug in casting the type
                    If Array.IndexOf(DpNotAllowed, CType(Stage.ToArray()(Stage.Count - 1), k)) >= 0 Then
                        ' Note: The real calculator is very forgiving (and silent) under
                        ' these conditions.  It just processes the decimal point as if
                        ' there was no staging.
                        Stage.Clear()
                    Else
                        Stage.Push(k.KeyDp)

                        ' return to the first stage for processing
                        Return ProcessKey(Stage.ToArray()(Stage.Count - 1))
                    End If
                End If

                ' Is this valid for the current mode?
                If cs.OpMode <> CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(key), ".")
                Else
                    If RawDisplay.Length = 0 Then
                        RawDisplay.Append("0.")
                        ' v6.0.4 - 29 May 12
                        ' Since we have artificially padded the display we need
                        ' another way to indicate that this is a new digit
                        PadDecimal = True
                    Else
                        ' extra decimal points are silently ignored
                        If RawDisplay.ToString.Contains(".") = False Then
                            RawDisplay.Append(".")
                        End If
                    End If
                End If
            Case k.KeyCHS           ' Change sign
                If PrgmEntry Then
                    recordLine(CInt(key), "CHS")
                Else
                    ' an exception to the rule for terminating input (page 99)
                    If RawDisplay.Length > 0 And cs.OpMode <> CalcState.CalcOpMode.Float Then
                        RawDisplay.Length = 0
                    End If

                    ' changing signs while keying digits
                    If RawDisplay.Length > 0 Then
                        Dim t As String = RawDisplay.ToString
                        RawDisplay.Length = 0
                        ' is there an exponent showing?
                        If t.Contains("e") Then
                            ' it's rather odd that once there is an exponent
                            ' you can no longer change the sign of the mantissa
                            If t.Contains("e-") Then
                                t = t.Replace("e-", "e")
                            Else
                                t = t.Replace("e", "e-")
                            End If
                            RawDisplay.Append(t)
                        Else
                            If t.StartsWith("-") Then
                                RawDisplay.Append(t.Substring(1))
                            Else
                                RawDisplay.Append("-")
                                RawDisplay.Append(t)
                            End If
                        End If
                    Else
                        ' just change the sign of what's in the X Register
                        If cs.OpMode = CalcState.CalcOpMode.Float Then
                            cs.Stack.X.FVal = -cs.Stack.X.FVal
                        Else
                            ' LastX is updated only in the integer mode (page 100)
                            cs.RegLastX = cs.Stack.X.Copy
                            cs.Stack.X.BiVal.ChangeSign()
                            cs.Flags(CalcState.CalcFlag.Overflow) = cs.Stack.X.BiVal.Overflow
                            Packet.MenuNeedsUpdating = True
                        End If
                        StackDisable = False
                    End If
                End If
            Case k.KeyAdd           ' Addition
                If PrgmEntry Then
                    recordLine(CInt(key), "+")
                Else
                    Dim temp As New Register(cs.WordSize, cs.ArithMode)
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        temp.FVal = cs.Stack.Pop.FVal + cs.Stack.Pop.FVal
                        cs.Flags(CalcState.CalcFlag.Overflow) = Double.IsInfinity(temp.FVal)
                    Else
                        temp.BiVal = BigInt.Add(cs.Stack.Pop.BiVal, cs.Stack.Pop.BiVal)
                        cs.Flags(CalcState.CalcFlag.Carry) = temp.BiVal.CarryBit
                        cs.Flags(CalcState.CalcFlag.Overflow) = temp.BiVal.Overflow
                    End If
                    Packet.MenuNeedsUpdating = True
                    cs.Stack.Push(temp)
                End If
            Case k.FnSL            ' Shift left
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnSL - F_KEY, "f SL")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    cs.Stack.X.BiVal.LeftShift(1, cs.WordSize)
                    cs.Flags(CalcState.CalcFlag.Carry) = cs.Stack.X.BiVal.CarryBit
                    Packet.MenuNeedsUpdating = True
                End If
            Case k.FnSR            ' Shift Right
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnSR - F_KEY, "f SR")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    cs.Stack.X.BiVal.RightShift(1, False, cs.WordSize)
                    cs.Flags(CalcState.CalcFlag.Carry) = cs.Stack.X.BiVal.CarryBit
                    Packet.MenuNeedsUpdating = True
                End If
            Case k.FnRL            ' Rotate Left
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnRL - F_KEY, "f RL")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    cs.Stack.X.BiVal.LeftRotate(1, False, False, cs.WordSize)
                    cs.Flags(CalcState.CalcFlag.Carry) = cs.Stack.X.BiVal.CarryBit
                    Packet.MenuNeedsUpdating = True
                End If
            Case k.FnRR            ' Rotate Right
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnRR - F_KEY, "f RR")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    cs.Stack.X.BiVal.RightRotate(1, False, False, cs.WordSize)
                    cs.Flags(CalcState.CalcFlag.Carry) = cs.Stack.X.BiVal.CarryBit
                    Packet.MenuNeedsUpdating = True
                End If
            Case k.FnRLn           ' Rotate Left n times
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnRLn - F_KEY, "f RLn")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    val = BigInt.AbsoluteValue(cs.Stack.Pop.BiVal).ToInteger
                    If val > cs.WordSize Then
                        Packet.AlternateText = "Error 2 - Improper Bit Number"
                        Prefix = 0
                        Return Packet
                    End If
                    cs.Stack.X.BiVal.LeftRotate(val, False, False, cs.WordSize)
                    cs.Flags(CalcState.CalcFlag.Carry) = cs.Stack.X.BiVal.CarryBit
                    Packet.MenuNeedsUpdating = True
                End If
            Case k.FnRRn           ' Rotate Right n times
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnRRn - F_KEY, "f RRn")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    val = BigInt.AbsoluteValue(cs.Stack.Pop.BiVal).ToInteger
                    If val > cs.WordSize Then
                        Packet.AlternateText = "Error 2 - Improper Bit Number"
                        Prefix = 0
                        Return Packet
                    End If
                    cs.Stack.X.BiVal.RightRotate(val, False, False, cs.WordSize)
                    cs.Flags(CalcState.CalcFlag.Carry) = cs.Stack.X.BiVal.CarryBit
                    Packet.MenuNeedsUpdating = True
                End If
            Case k.FnMASKL         ' Create Left justified Mask
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnMASKL - F_KEY, "f MASKL")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    Dim temp As New Register(cs.WordSize, cs.ArithMode)
                    val = BigInt.AbsoluteValue(cs.Stack.Pop().BiVal).ToInteger()
                    If val > cs.WordSize Then
                        Packet.AlternateText = "Error 2 - Improper Bit Number"
                        Prefix = 0
                        Return Packet
                    End If
                    temp.BiVal = BigInt.CreateMask(val, True, cs.WordSize, cs.ArithMode)
                    cs.Stack.Push(temp)
                End If
            Case k.FnMASKR         ' Create Right justified Mask
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnMASKR - F_KEY, "f MASKR")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    Dim temp As New Register(cs.WordSize, cs.ArithMode)
                    val = BigInt.AbsoluteValue(cs.Stack.Pop().BiVal).ToInteger()
                    If val > cs.WordSize Then
                        Packet.AlternateText = "Error 2 - Improper Bit Number"
                        Prefix = 0
                        Return Packet
                    End If
                    temp.BiVal = BigInt.CreateMask(val, False, cs.WordSize, cs.ArithMode)
                    cs.Stack.Push(temp)
                End If
            Case k.FnRMD           ' Remainder after division
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnRMD - F_KEY, "f RMD")
                Else
                    Dim temp As New Register(cs.WordSize, cs.ArithMode)
                    Dim y As BigInt

                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    ' v6.0.7 - 22 Jul 2015
                    If BigInt.AbsoluteValue(cs.Stack.X.BiVal).IsZero Then
                        Packet.AlternateText = "Error 0 - Improper Math Operation"
                        Prefix = 0
                        Return Packet
                    End If
                    y = cs.Stack.Pop.BiVal
                    temp.BiVal = BigInt.Remainder(cs.Stack.Pop.BiVal, y)
                    cs.Stack.Push(temp)
                End If
            Case k.FnXOR           ' Bitwise Exclusive OR
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnXOR - F_KEY, "f XOR")
                Else
                    Dim temp As New Register
                    Dim y As BigInt

                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    y = cs.Stack.Pop.BiVal
                    temp.BiVal = BigInt.[Xor](cs.Stack.Pop.BiVal, y)
                    cs.Stack.Push(temp)
                End If
            Case k.FnXIndex        ' Swap X with indirect value of I
                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnXIndex - F_KEY, "f X:(i)")
                Else
                    Dim temp As Register
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        ' The absolute value of the integer portion
                        val = CInt(Math.Abs(cs.RegIndex.FVal))
                    Else
                        val = BigInt.AbsoluteValue(cs.RegIndex.BiVal).ToInteger
                    End If
                    StackDisable = False

                    ' a quick sanity check
                    If val >= numRegs Then
                        Packet.AlternateText = "Error 3 - Improper Register Number"
                        Prefix = 0
                        Return Packet
                    End If

                    temp = cs.Reg(val).Copy
                    cs.Reg(val) = cs.Stack.X.Copy
                    cs.Stack.X = temp
                End If
            Case k.FnXI            ' Swap X and I
                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnXI - F_KEY, "f X:I")
                Else
                    Dim temp As Register
                    temp = cs.RegIndex.Copy
                    temp.BiVal.WordSize = cs.WordSize
                    cs.RegIndex = cs.Stack.X.Copy
                    cs.RegIndex.BiVal.WordSize = 64
                    cs.Stack.X = temp
                    StackDisable = False
                End If
            Case k.FnShowHex       ' Show HEX
                If syncConv Then
                    ' Note: The real calculator doesn't allow this in the float mode
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        ' synchronize the float and integer values
                        cs.SyncValues()
                    End If
                Else
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        Packet.Beep = True
                        Exit Select
                    End If
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnShowHex - F_KEY, "f ShowHEX")
                Else
                    Prefix = 0
                    Packet.AlternateText = FormatDisplay(CalcState.CalcOpMode.Hex, 0)
                    Packet.Delay = sleepDelay
                    Return Packet
                End If
            Case k.FnShowDec       ' Show DEC
                If syncConv Then
                    ' Note: The real calculator doesn't allow this in the float mode
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        ' synchronize the float and integer values
                        cs.SyncValues()
                    End If
                Else
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        Packet.Beep = True
                        Exit Select
                    End If
                End If
                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnShowDec - F_KEY, "f ShowDEC")
                Else
                    Prefix = 0
                    Packet.AlternateText = FormatDisplay(CalcState.CalcOpMode.Dec, 0)
                    Packet.Delay = sleepDelay
                    Return Packet
                End If
            Case k.FnShowOct       ' Show OCT
                If syncConv Then
                    ' Note: The real calculator doesn't allow this in the float mode
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        ' synchronize the float and integer values
                        cs.SyncValues()
                    End If
                Else
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        Packet.Beep = True
                        Exit Select
                    End If
                End If
                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnShowOct - F_KEY, "f ShowOCT")
                Else
                    Prefix = 0
                    Packet.AlternateText = FormatDisplay(CalcState.CalcOpMode.Oct, 0)
                    Packet.Delay = sleepDelay
                    Return Packet
                End If
            Case k.FnShowBin       ' Show BIN
                If syncConv Then
                    ' Note: The real calculator doesn't allow this in the float mode
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        ' synchronize the float and integer values
                        cs.SyncValues()
                    End If
                Else
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        Packet.Beep = True
                        Exit Select
                    End If
                End If
                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnShowBin - F_KEY, "f ShowBIN")
                Else
                    Prefix = 0
                    Packet.AlternateText = FormatDisplay(CalcState.CalcOpMode.Bin, 0)
                    Packet.Delay = sleepDelay
                    Return Packet
                End If
            Case k.FnSB            ' Set Bit  
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If
                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnSB - F_KEY, "f SB")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    val = BigInt.AbsoluteValue(cs.Stack.Pop.BiVal).ToInteger
                    If val >= cs.WordSize Then
                        Packet.AlternateText = "Error 2 - Improper Bit Number"
                        Prefix = 0
                        Return Packet
                    End If
                    cs.Stack.X.BiVal.BitSet(val)
                End If
            Case k.FnCB            ' Clear Bit
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If
                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnCB - F_KEY, "f CB")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    val = BigInt.AbsoluteValue(cs.Stack.Pop.BiVal).ToInteger
                    If val >= cs.WordSize Then
                        Packet.AlternateText = "Error 2 - Improper Bit Number"
                        Prefix = 0
                        Return Packet
                    End If
                    cs.Stack.X.BiVal.BitClear(val)
                End If
            Case k.FnBSet          ' Is the bit set?
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If
                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnBSet - F_KEY, "f B?")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    val = BigInt.AbsoluteValue(cs.Stack.Pop.BiVal).ToInteger
                    If val >= cs.WordSize Then
                        Packet.AlternateText = "Error 2 - Improper Bit Number"
                        Prefix = 0
                        Return Packet
                    End If
                    If Not cs.Stack.X.BiVal.BitTest(val) Then
                        cs.PrgmPosition += 1
                    End If
                End If
            Case k.FnAND           ' Bitwise AND operator
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If
                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnAND - F_KEY, "f AND")
                Else
                    Dim temp As New Register
                    Dim y As BigInt

                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    y = cs.Stack.Pop.BiVal
                    temp.BiVal = BigInt.BitwiseAnd(cs.Stack.Pop.BiVal, y)
                    cs.Stack.Push(temp)
                End If
            Case k.FnIndex         ' Indirect value of I
                ' OK, this one is a bit odd... if there is no stage, then
                ' act like a RCL was staged. (Note: This is not in the book)
                If Stage.Count = 0 Then
                    Stage.Push(k.KeyRCL)
                End If

                Stage.Push(k.FnIndex)

                ' return to the first stage for processing
                Prefix = 0
                Return ProcessKey(Stage.ToArray()(Stage.Count - 1))
                ' Note: No PrgmMode processing is required here, since this is always a stage operand
            Case k.FnI             ' Index
                ' OK, this one is a bit odd... if there is no stage, then
                ' act like a RCL was staged. (Note: This is not in the book)
                If Stage.Count = 0 Then
                    Stage.Push(k.KeyRCL)
                End If

                Stage.Push(k.FnI)

                ' return to the first stage for processing
                Prefix = 0
                Return ProcessKey(Stage.ToArray()(Stage.Count - 1))
                ' Note: No PrgmMode processing is required here, since this is always a stage operand
            Case k.FnClearPrgm     ' Clear Program
                ' This has two different meanings... if in the program
                ' mode it will clear all program lines.  However, if in
                ' the Run mode, it will merely reset the program position
                ' to zero (the program is NOT erased)
                If PrgmEntry Then
                    cs.PrgmMemory.Clear()
                    cs.PrgmRetStack.Clear()
                End If
                StackDisable = False
                cs.PrgmPosition = 0
            Case k.FnClearReg      ' Clear all storage registers
                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnClearReg - F_KEY, "f ClearREG")
                Else
                    ' Clears storage registers (and Index), but not the stack
                    For i As Integer = 0 To numRegs - 1
                        cs.Reg(i) = New Register(cs.WordSize, cs.ArithMode)
                    Next
                    cs.RegIndex = New Register(64, cs.ArithMode)
                End If
            Case k.FnClearPrefix   ' Clear any pending prefix/stage
                Prefix = 0
                Stage.Clear()
                RawDisplay.Append(previousRaw)
                Return Packet
                ' Note: No PrgmMode, since this is just used for editing
            Case k.FnWINDOW        ' show a window
                ' This is a multi-stage command, so we have to store 1
                ' more keystroke before we can do anything
                If Stage.Count = 0 Then
                    Stage.Push(k.FnWINDOW)
                    Exit Select
                End If

                ' a quick sanity check
                If Stage.ToArray()(Stage.Count - 1) <> k.FnWINDOW Then
                    Stage.Clear()
                    Stage.Push(k.FnWINDOW)
                    Exit Select
                End If

                ' Since our display can hold 32 bits, we only need two
                ' windows (rather than 8 on the real calculator)
                val = Stage.Pop
                If val < 0 Or val > 7 Then
                    Stage.Clear()
                    Packet.AlternateText = "Error 1 - Improper Windows Number"
                    Prefix = 0
                    Return Packet
                End If

                Stage.Clear()

                ' we'll cut you some slack... for old time's sake
                If val > 1 Then
                    val = 1
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnWINDOW - F_KEY, val, "f WINDOW " & val)
                End If
                ' only allow this when the conditions warrant it
                If cs.OpMode = CalcState.CalcOpMode.Bin And cs.WordSize > 32 Then
                    ' this is just a code for Window 1
                    WinPos = val * 36
                End If
            Case k.FnSet1s         ' switch to 1's complement mode
                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnSet1s - F_KEY, "f Set1's")
                Else
                    cs.ArithMode = BigInt.ArithMode.OnesComp
                    Packet.MenuNeedsUpdating = True
                End If
            Case k.FnSet2s         ' switch to 2's complement mode
                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnSet2s - F_KEY, "f Set2's")
                Else
                    cs.ArithMode = BigInt.ArithMode.TwosComp
                    Packet.MenuNeedsUpdating = True
                End If
            Case k.FnSetUnsgn      ' switch to unsigned mode
                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnSetUnsgn - F_KEY, "f SetUNSGN")
                Else
                    cs.ArithMode = BigInt.ArithMode.Unsigned
                    Packet.MenuNeedsUpdating = True
                End If
            Case k.FnNOT           ' The Bitwise NOT operator
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnNOT - F_KEY, "f NOT")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    cs.Stack.X.BiVal.OnesComplement(cs.WordSize)
                End If
            Case k.FnWSIZE         ' set the word size
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnWSIZE - F_KEY, "f WSIZE")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    val = BigInt.AbsoluteValue(cs.Stack.Pop.BiVal).ToInteger

                    ' This is the "escape clause", allowing you to select a new word size
                    ' when the current word size won't allow the new larger number to be
                    ' entered.
                    If val = 0 Then
                        val = 64
                    End If

                    ' a quick sanity check
                    If val > 64 Then
                        Packet.AlternateText = "Error 2 - Improper Bit Number"
                        Prefix = 0
                        Return Packet
                    End If
                    cs.WordSize = val
                    Packet.MenuNeedsUpdating = True
                End If
            Case k.FnFLOAT         ' set the float mode
                ' This is a multi-stage command, so we have to store 1
                ' more keystroke before we can do anything
                If Stage.Count = 0 Then
                    Stage.Push(k.FnFLOAT)
                    Exit Select
                End If

                ' a quick sanity check
                If Stage.ToArray()(Stage.Count - 1) <> k.FnFLOAT Then
                    Stage.Clear()
                    Stage.Push(k.FnFLOAT)
                    Exit Select
                End If

                ' Check for valid stage operands
                val = Stage.Pop
                If val <> k.KeyDp And val > 9 Then
                    Stage.Clear()
                    Packet.AlternateText = "Error 1 - Improper Float Number"
                    Prefix = 0
                    Return Packet
                End If

                Stage.Clear()

                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnFLOAT - F_KEY, val, "f FLOAT " & val)
                Else
                    If cs.OpMode <> CalcState.CalcOpMode.Float Then
                        ' Here is a strange LastX rule (page 100)
                        cs.RegLastX = New Register(cs.WordSize, cs.ArithMode)
                        cs.SyncValues()

                        ' reset the carry and overflow flags
                        cs.Flags(CalcState.CalcFlag.Carry) = False
                        cs.Flags(CalcState.CalcFlag.Overflow) = False
                    Else
                        ' only in float mode
                        StackDisable = False
                    End If
                    cs.OpMode = CalcState.CalcOpMode.Float
                    cs.FloatPrecision = val
                    Packet.MenuNeedsUpdating = True
                End If
            Case k.FnMEM           ' Display the Memory configuration
                ' It should be obvious that these values don't change
                Packet.AlternateText = String.Format("P-{0:000} R-{1:000}", numLines, numRegs)
                Packet.Delay = sleepDelay
                Prefix = 0
                Return Packet
                ' Note: MEM can not be stored in Program Mode
            Case k.FnSTATUS        ' Display the Calculator configuration
                ' Combine flags 0-3 into a single number
                Dim FlagVal As Integer = 0
                For i As Integer = 0 To 3
                    FlagVal += If(cs.Flags(i), CInt(Math.Pow(10, i)), 0)
                Next

                Packet.Delay = sleepDelay
                ' v6.0.3 - 5 May 12
                Packet.AlternateText = String.Format("{0:0}-{1:00}-{2:0000}", CInt(cs.ArithMode), cs.WordSize, FlagVal)
                Prefix = 0
                Return Packet
                ' Note: STATUS can not be stored in Program Mode
            Case k.FnEEX           ' Enter exponent
                ' Is this valid for the current mode?
                If cs.OpMode <> CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnEEX - F_KEY, "f EEX")
                Else
                    ' This is a bit strange... if this is the first keystroke,
                    ' then let's insert a 1 before it
                    If RawDisplay.Length = 0 Then
                        RawDisplay.Append("1")
                    End If
                    RawDisplay.Append("e")
                End If
            Case k.FnOR            ' The Bitwise OR operator
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyFKey), k.FnOR - F_KEY, "f OR")
                Else
                    Dim temp As New Register
                    Dim y As BigInt

                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    y = cs.Stack.Pop.BiVal
                    temp.BiVal = BigInt.BitwiseOr(cs.Stack.Pop.BiVal, y)
                    cs.Stack.Push(temp)
                End If
            Case k.FnLJ            ' Left Justify
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnLJ - G_KEY, "g LJ")
                Else
                    Dim distance As New Register(cs.WordSize, cs.ArithMode)
                    Dim justified As New Register(cs.WordSize, cs.ArithMode)

                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    Dim ret As BigInt() = BigInt.LeftJustify(cs.Stack.Pop.BiVal)
                    distance.BiVal = ret(0)
                    justified.BiVal = ret(1)

                    cs.Stack.Push(distance)
                    cs.Stack.Push(justified)
                End If
            Case k.FnASR           ' Arithmetic Shift
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnASR - G_KEY, "g ASR")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    cs.Stack.X.BiVal.RightShift(1, True, cs.WordSize)
                    cs.Flags(CalcState.CalcFlag.Carry) = cs.Stack.X.BiVal.CarryBit
                    Packet.MenuNeedsUpdating = True
                End If
            Case k.FnRLC           ' Rotate Left with Carry
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnRLC - G_KEY, "g RLC")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    cs.Stack.X.BiVal.LeftRotate(1, True, cs.Flags(CalcState.CalcFlag.Carry), cs.WordSize)
                    cs.Flags(CalcState.CalcFlag.Carry) = cs.Stack.X.BiVal.CarryBit
                    Packet.MenuNeedsUpdating = True
                End If
            Case k.FnRRC           ' Rotate Right with Carry
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnRRC - G_KEY, "g RRC")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    cs.Stack.X.BiVal.RightRotate(1, True, cs.Flags(CalcState.CalcFlag.Carry), cs.WordSize)
                    cs.Flags(CalcState.CalcFlag.Carry) = cs.Stack.X.BiVal.CarryBit
                    Packet.MenuNeedsUpdating = True
                End If
            Case k.FnRLCn         ' Rotate Left with Carry n times
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnRLCn - G_KEY, "g RLCn")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    val = BigInt.AbsoluteValue(cs.Stack.Pop.BiVal).ToInteger
                    If val > cs.WordSize Then
                        Packet.AlternateText = "Error 2 - Improper Bit Number"
                        Prefix = 0
                        Return Packet
                    End If
                    cs.Stack.X.BiVal.LeftRotate(val, True, cs.Flags(CalcState.CalcFlag.Carry), cs.WordSize)
                    cs.Flags(CalcState.CalcFlag.Carry) = cs.Stack.X.BiVal.CarryBit
                    Packet.MenuNeedsUpdating = True
                End If
            Case k.FnRRCn         ' Rotate Right with Carry n times
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnRRCn - G_KEY, "g RRCn")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    val = BigInt.AbsoluteValue(cs.Stack.Pop.BiVal).ToInteger
                    If val > cs.WordSize Then
                        Packet.AlternateText = "Error 2 - Improper Bit Number"
                        Prefix = 0
                        Return Packet
                    End If
                    cs.Stack.X.BiVal.RightRotate(val, True, cs.Flags(CalcState.CalcFlag.Carry), cs.WordSize)
                    cs.Flags(CalcState.CalcFlag.Carry) = cs.Stack.X.BiVal.CarryBit
                    Packet.MenuNeedsUpdating = True
                End If
            Case k.FnNumB         ' Number of bits
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnNumB - G_KEY, "g #B")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    cs.Stack.X.BiVal.SumBits()
                End If
            Case k.FnABS         ' Absolute value
                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnABS - G_KEY, "g ABS")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        cs.Stack.X.FVal = Math.Abs(cs.Stack.X.FVal)
                    Else
                        cs.Stack.X.BiVal.AbsoluteValue()
                        cs.Flags(CalcState.CalcFlag.Carry) = cs.Stack.X.BiVal.CarryBit
                        Packet.MenuNeedsUpdating = True
                    End If
                End If
            Case k.FnDBLR        ' Double remainder (after division)
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnDBLR - G_KEY, "g DBLR")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    If BigInt.AbsoluteValue(cs.Stack.X.BiVal).IsZero Then
                        Packet.AlternateText = "Error 0 - Improper Math Operation"
                        Prefix = 0
                        Return Packet
                    End If

                    Dim x, y, z As Register
                    ' v7.0.0 - 6 Jun 21: Fixed Double funtions
                    Dim large As New Register(cs.WordSize, cs.ArithMode)
                    x = cs.Stack.Pop
                    y = cs.Stack.Pop
                    z = cs.Stack.Pop
                    large.BiVal.Combine(y.BiVal, z.BiVal)
                    large.BiVal = BigInt.Remainder(large.BiVal, x.BiVal)
                    If large.BiVal.Overflow Then
                        Packet.AlternateText = "Error 0 - Improper Math Operation"
                        Prefix = 0
                        Return Packet
                    End If
                    large.BiVal.WordSize = cs.WordSize
                    cs.Stack.Push(large)
                End If
            Case k.FnDBLDiv      ' Double division
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnDBLR - G_KEY, "g DBL/")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    If BigInt.AbsoluteValue(cs.Stack.X.BiVal).IsZero Then
                        Packet.AlternateText = "Error 0 - Improper Math Operation"
                        Prefix = 0
                        Return Packet
                    End If
                    Dim x, y, z, temp2 As Register
                    ' v7.0.0 - 6 Jun 21: Fixed Double funtions
                    Dim temp As New Register(cs.WordSize, cs.ArithMode)
                    x = cs.Stack.Pop
                    y = cs.Stack.Pop
                    z = cs.Stack.Pop
                    temp.BiVal.Combine(y.BiVal, z.BiVal)
                    temp.BiVal.Divide(x.BiVal, cs.WordSize)

                    ' does the answer fit?
                    temp2 = temp.Copy
                    temp2.BiVal.WordSize = cs.WordSize
                    If Not temp2.BiVal.Equals(temp.BiVal) Then
                        ' if not, then set error zero (page 53)
                        Packet.AlternateText = "Error 0 - Improper Math Operation"
                        Prefix = 0
                        Return Packet
                    End If

                    cs.Stack.Push(temp2)
                    cs.Flags(CalcState.CalcFlag.Carry) = temp.BiVal.CarryBit

                    ' always clears the overflow flag (page 98)
                    cs.Flags(CalcState.CalcFlag.Overflow) = False
                    Packet.MenuNeedsUpdating = True
                End If
            Case k.FnRTN         ' Return from subroutine (Program Mode)
                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnRTN - G_KEY, "g RTN")
                Else
                    ' pop the return position from the subroutine stack
                    If cs.PrgmRetStack.Count = 0 Then
                        ' an "empty" return also means stop
                        cs.PrgmRunning = False
                    Else
                        cs.PrgmPosition = cs.PrgmRetStack.Pop
                    End If
                End If
            Case k.FnLBL         ' A program Label (Program Mode)
                ' This is a multi-stage command, so we have to store 1
                ' more keystroke before we can do anything
                If Stage.Count = 0 Then
                    Stage.Push(k.FnLBL)
                    Exit Select
                End If

                ' a quick sanity check
                If Stage.ToArray()(Stage.Count - 1) <> k.FnLBL Then
                    Stage.Clear()
                    Stage.Push(k.FnLBL)
                    Exit Select
                End If

                ' Get the stage operand
                val = Stage.Pop
                Stage.Clear()

                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnLBL - G_KEY, val, "g LBL " & Hex(val))
                End If
            Case k.FnDSZ         ' Decrement, skip on zero (Program Mode)
                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnDSZ - G_KEY, "g DSZ")
                Else
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        cs.RegIndex.FVal -= 1
                        If cs.RegIndex.FVal = 0 Then
                            cs.PrgmPosition += 1
                        End If
                    Else
                        cs.RegIndex.BiVal.Subtract(BigInt.One(cs.WordSize), 64)
                        If BigInt.AbsoluteValue(cs.RegIndex.BiVal).IsZero Then
                            cs.PrgmPosition += 1
                        End If
                    End If
                End If
            Case k.FnISZ         ' Increment, skip on zero (Program Mode)
                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnISZ - G_KEY, "g ISZ")
                Else
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        cs.RegIndex.FVal += 1
                        If cs.RegIndex.FVal = 0 Then
                            cs.PrgmPosition += 1
                        End If
                    Else
                        cs.RegIndex.BiVal.Add(BigInt.One(cs.WordSize), 64)
                        ' v6.0.8 - 11 Jan 2016 (thanks Andru)
                        If BigInt.AbsoluteValue(cs.RegIndex.BiVal).IsZero Then
                            cs.PrgmPosition += 1
                        End If
                    End If
                End If
            Case k.FnSqrt        ' Square root
                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnSqrt - G_KEY, "g Sqrt")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        ' v7.0.0 - 6 Jun 21: Thanks Andre Hut
                        If cs.Stack.X.FVal < 0.0 Then
                            Packet.AlternateText = "Error 0 - Improper Math Operation"
                            Prefix = 0
                            Return Packet
                        End If
                        cs.Stack.X.FVal = Math.Sqrt(cs.Stack.X.FVal)
                    Else
                        Dim zero As New BigInt(cs.Stack.X.BiVal.WordSize, cs.Stack.X.BiVal.BIArithMode)
                        If cs.Stack.X.BiVal.IsLess(zero) Then
                            Packet.AlternateText = "Error 0 - Improper Math Operation"
                            Prefix = 0
                            Return Packet
                        End If
                        cs.Stack.X.BiVal.SquareRoot(cs.WordSize)
                        cs.Flags(CalcState.CalcFlag.Carry) = cs.Stack.X.BiVal.CarryBit
                        Packet.MenuNeedsUpdating = True
                    End If
                End If
            Case k.FnInv         ' Inverse
                ' Is this valid for the current mode?
                If cs.OpMode <> CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnInv - G_KEY, "g 1/x")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    If cs.Stack.X.FVal = 0.0 Then
                        Packet.AlternateText = "Error 0 - Improper Math Operation"
                        Prefix = 0
                        Return Packet
                    End If
                    cs.Stack.X.FVal = 1.0 / cs.Stack.X.FVal
                End If
            Case k.FnSF            ' Set Flag
                ' This is a multi-stage command, so we have to store 1
                ' more keystroke before we can do anything
                If Stage.Count = 0 Then
                    Stage.Push(k.FnSF)
                    Exit Select
                End If

                ' a quick sanity check
                If Stage.ToArray()(Stage.Count - 1) <> k.FnSF Then
                    Stage.Clear()
                    Stage.Push(k.FnSF)
                    Exit Select
                End If
                StackDisable = False

                ' Check for valid stage operand
                val = Stage.Pop
                If val < 0 Or val > 5 Then
                    Stage.Clear()
                    Packet.AlternateText = "Error 1 - Improper Flag Number"
                    Prefix = 0
                    Return Packet
                End If

                Stage.Clear()

                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnSF - G_KEY, val, "g SF " & val)
                Else
                    cs.Flags(val) = True
                    Packet.MenuNeedsUpdating = True
                End If
            Case k.FnCF            ' Clear Flag
                ' This is a multi-stage command, so we have to store 1
                ' more keystroke before we can do anything
                If Stage.Count = 0 Then
                    Stage.Push(k.FnCF)
                    Exit Select
                End If

                ' a quick sanity check
                If Stage.ToArray()(Stage.Count - 1) <> k.FnCF Then
                    Stage.Clear()
                    Stage.Push(k.FnCF)
                    Exit Select
                End If
                StackDisable = False

                ' Check for valid stage operand
                val = Stage.Pop
                If val > 5 Then
                    Stage.Clear()
                    Packet.AlternateText = "Error 1 - Improper Flag Number"
                    Prefix = 0
                    Return Packet
                End If

                Stage.Clear()

                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnCF - G_KEY, val, "g CF " & val)
                Else
                    cs.Flags(val) = False
                    Packet.MenuNeedsUpdating = True
                End If
            Case k.FnFSet            ' Is the Flag set
                ' This is a multi-stage command, so we have to store 1
                ' more keystroke before we can do anything
                If Stage.Count = 0 Then
                    Stage.Push(k.FnFSet)
                    Exit Select
                End If

                ' a quick sanity check
                If Stage.ToArray()(Stage.Count - 1) <> k.FnFSet Then
                    Stage.Clear()
                    Stage.Push(k.FnFSet)
                    Exit Select
                End If

                ' Check for valid stage operand
                val = Stage.Pop
                If val > 5 Then
                    Stage.Clear()
                    Packet.AlternateText = "Error 1 - Improper Flag Number"
                    Prefix = 0
                    Return Packet
                End If

                Stage.Clear()

                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnFSet - G_KEY, val, "g F? " & val)
                Else
                    If cs.Flags(val) = False Then
                        cs.PrgmPosition += 1
                    End If
                End If
            Case k.FnDBLMul        ' Double multiplication
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnDBLMul - G_KEY, "g DBL*")
                Else
                    cs.RegLastX = cs.Stack.X.Copy
                    StackDisable = False
                    Dim large_y As Register
                    Dim x As New Register(cs.WordSize, cs.ArithMode)
                    Dim y As New Register(cs.WordSize, cs.ArithMode)
                    large_y = cs.Stack.Pop
                    large_y.BiVal.WordSize = cs.WordSize * 2
                    large_y.BiVal.Multiply(cs.Stack.Pop.BiVal, cs.WordSize * 2)
                    Dim ret() As BigInt = large_y.BiVal.Split
                    x.BiVal = ret(0)
                    y.BiVal = ret(1)
                    cs.Stack.Push(x)
                    cs.Stack.Push(y)

                    ' always clears overflow (page 98)
                    cs.Flags(CalcState.CalcFlag.Overflow) = False
                    Packet.MenuNeedsUpdating = True
                End If
            Case k.FnPR            ' Toggle between Program and Run mode
                PrgmEntry = Not PrgmEntry
                Packet.Prgm_Annunciator = PrgmEntry
            Case k.FnBST           ' Back step (Program Mode)
                If PrgmEntry Then
                    If cs.PrgmPosition > 0 Then
                        cs.PrgmPosition -= 1
                    Else
                        cs.PrgmPosition = cs.PrgmMemory.Count
                    End If
                Else
                    ' v7.0.0 - 6 Jun 21: Forgot to do this...
                    If cs.PrgmMemory.Count > 0 Then
                        If cs.PrgmPosition <= 0 Then
                            cs.PrgmPosition = cs.PrgmMemory.Count
                        End If
                        cs.PrgmPosition -= 1
                    End If
                    ' temporarily show the program codes (Note: Does Not execute the line)
                    Packet.AlternateText = (cs.PrgmPosition + 1).ToString("000") + "- " + cs.PrgmMemory(cs.PrgmPosition).Substring(0, 8)
                    Packet.Delay = CInt(sleepDelay / 1.5)
                End If
                ' Note: BST can not be stored
            Case k.FnRolUp         ' Roll the stack up
                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnRolUp - G_KEY, "g R^")
                Else
                    Dim temp As Register
                    temp = cs.Stack.T.Copy
                    cs.Stack.T = cs.Stack.Z.Copy
                    cs.Stack.Z = cs.Stack.Y.Copy
                    cs.Stack.Y = cs.Stack.X.Copy
                    cs.Stack.X = temp
                    StackDisable = False
                End If
            Case k.FnPSE           ' Pause execution (Program Mode)
                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnPSE - G_KEY, "g PSE")
                Else
                    Packet.DisplayText = FormatDisplay()
                    Packet.Delay = sleepDelay
                End If
            Case k.FnCLX           ' Clear the X register
                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnCLX - G_KEY, "g CLx")
                Else
                    cs.Stack.X = New Register(cs.WordSize, cs.ArithMode)
                    ' "disable" the stack for the next entry
                    StackDisable = True
                End If
            Case k.FnLSTX          ' Recall the contents of Last X register
                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnLSTX - G_KEY, "g LSTx")
                Else
                    ' v7.0.0 - 6 Jun 21: thanks Neil Singer
                    cs.Stack.Push(cs.RegLastX.Copy)
                    StackDisable = False
                End If
            Case k.FnXlteY         ' Is X less than or equal to Y (Program Mode)
                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnXlteY - G_KEY, "g X<Y")
                Else
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        If cs.Stack.X.FVal > cs.Stack.Y.FVal Then
                            cs.PrgmPosition += 1
                        End If
                    Else
                        If cs.Stack.X.BiVal.IsGreater(cs.Stack.Y.BiVal) Then
                            cs.PrgmPosition += 1
                        End If
                    End If
                End If
            Case k.FnXlt0          ' Is X less than 0 (Program Mode)
                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnXlt0 - G_KEY, "g X<0")
                Else
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        If cs.Stack.X.FVal >= 0.0 Then
                            cs.PrgmPosition += 1
                        End If
                    Else
                        If cs.Stack.X.BiVal.ToLong >= 0L Then
                            cs.PrgmPosition += 1
                        End If
                    End If
                End If
            Case k.FnXgtY          ' Is X greater than Y (Program Mode)
                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnXgtY - G_KEY, "g X>Y")
                Else
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        If cs.Stack.X.FVal <= cs.Stack.Y.FVal Then
                            cs.PrgmPosition += 1
                        End If
                    Else
                        If cs.Stack.X.BiVal.IsLessOrEqual(cs.Stack.Y.BiVal) Then
                            cs.PrgmPosition += 1
                        End If
                    End If
                End If
            Case k.FnXgt0          ' Is X greater than 0 (Program Mode)
                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnXgt0 - G_KEY, "g X>0")
                Else
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        If cs.Stack.X.FVal <= 0.0 Then
                            cs.PrgmPosition += 1
                        End If
                    Else
                        If cs.Stack.X.BiVal.ToLong <= 0L Then
                            cs.PrgmPosition += 1
                        End If
                    End If
                End If
            Case k.FnSLeft         ' Scroll the display Left
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnSLeft - G_KEY, "g <")
                Else
                    ' only allow this when the conditions warrant it
                    If cs.OpMode = CalcState.CalcOpMode.Bin And cs.WordSize > 32 Then
                        WinPos -= 1
                        If WinPos < 0 Then
                            WinPos = 0
                        End If
                    End If
                End If
            Case k.FnSRight        ' Scroll the display Right
                ' Is this valid for the current mode?
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    Packet.Beep = True
                    Exit Select
                End If

                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnSRight - G_KEY, "g >")
                Else
                    ' only allow this when the conditions warrant it
                    If cs.OpMode = CalcState.CalcOpMode.Bin And cs.WordSize > 32 Then
                        Dim len As Integer = cs.Stack.X.BiVal.ToStringBin(True).Length
                        WinPos += 1
                        If WinPos > len - 35 Then
                            ' this is just a code to indicate window 1
                            WinPos = len - 35
                        End If
                    End If
                End If
            Case k.FnXneY          ' Is X not equal to Y (Program Mode)
                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnXneY - G_KEY, "g X<>Y")
                Else
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        If cs.Stack.X.FVal = cs.Stack.Y.FVal Then
                            cs.PrgmPosition += 1
                        End If
                    Else
                        If cs.Stack.X.BiVal.Equals(cs.Stack.Y.BiVal) Then
                            cs.PrgmPosition += 1
                        End If
                    End If
                End If
            Case k.FnXne0          ' Is X not equal to 0 (Program Mode)
                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnXne0 - G_KEY, "g X=0")
                Else
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        If cs.Stack.X.FVal = 0.0 Then
                            cs.PrgmPosition += 1
                        End If
                    Else
                        If BigInt.AbsoluteValue(cs.Stack.X.BiVal).IsZero Then
                            cs.PrgmPosition += 1
                        End If
                    End If
                End If
            Case k.FnXeqY          ' Is X equal to Y (Program Mode)
                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnXeqY - G_KEY, "g X=Y")
                Else
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        If cs.Stack.X.FVal <> cs.Stack.Y.FVal Then
                            cs.PrgmPosition += 1
                        End If
                    Else
                        If Not cs.Stack.X.BiVal.Equals(cs.Stack.Y.BiVal) Then
                            cs.PrgmPosition += 1
                        End If
                    End If
                End If
            Case k.FnXeq0          ' Is X equal to 0 (Program Mode)
                If PrgmEntry Then
                    recordLine(CInt(k.KeyGKey), k.FnXeq0 - G_KEY, "g X=0")
                Else
                    If cs.OpMode = CalcState.CalcOpMode.Float Then
                        If cs.Stack.X.FVal <> 0.0 Then
                            cs.PrgmPosition += 1
                        End If
                    Else
                        If Not BigInt.AbsoluteValue(cs.Stack.X.BiVal).IsZero Then
                            cs.PrgmPosition += 1
                        End If
                    End If
                End If
            Case Else
        End Select

        ' There are three types of clean up operations...
        ' 1) Terminate an incomplete stage
        ' 2) Clear the prefix
        ' 3) Reset the display window to 0

        ' Terminate an incomplete stage (if any)
        If Stage.Count > 0 AndAlso Stage.ToArray()(Stage.Count - 1) <> key Then
            If key > 15 And key <> k.KeyDp And key <> k.FnI And key <> k.FnIndex Then
                ' The real calculator is very forgiving (and silent) when you
                ' enter a nonsensical stage operand (such as "STO +").  So, I 
                ' guess I'll do the same here (except for the silent part)
                Packet.Beep = True
                Stage.Clear()
            End If
        End If

        ' clear the prefix (if set)
        Prefix = 0

        ' should we reset the window position?
        If Array.IndexOf(DoNotResetWinPos, key) < 0 Then
            WinPos = 0
        End If

        ' There are 3 things that could be showing in the display
        ' 1) program lines during the "Program Entry" mode
        ' 2) Raw input before it has been formatted
        ' 3) Fully formatted register values

        If PrgmEntry Then
            ' Is this test even necessary?  The PrgmMemoryLines value is quite
            ' arbitrary.  I could just not bother, and nobody would notice.
            If cs.PrgmMemory.Count = numLines Then
                Packet.AlternateText = "Error 4 - Improper Line Number"
                cs.PrgmMemory.RemoveAt(numLines)
                cs.PrgmPosition = numLines - 1
                Return Packet
            End If

            ' v7.0.2 - 28 Jun 23
            If cs.PrgmPosition > 0 And cs.PrgmMemory.Count > 0 Then
                Packet.DisplayText = cs.PrgmPosition.ToString("000") & "- " & Left(cs.PrgmMemory(cs.PrgmPosition - 1), 8)
            Else
                Packet.DisplayText = "000-"
            End If
        Else
            ' Should we terminate any existing digit entry?  There are only
            ' a few keys that do NOT terminate entry (page 99)
            If key > 15 And Array.IndexOf(TerminateInput, key) < 0 Then
                previousRaw = RawDisplay.ToString
                RawDisplay.Length = 0
            End If

            ' Either display the raw text or show the value of the
            ' X register
            If RawDisplay.Length > 0 Then
                ' is this a new entry?  If so, then push the existing value of
                ' the X register deeper onto the stack.
                ' v6.0.4 - 29 May 12
                If (RawDisplay.Length = 1 Or PadDecimal) And Not StackDisable Then
                    cs.Stack.Push(cs.Stack.X)
                End If
                StackDisable = False
                PadDecimal = False

                ' put it in X, but don't (yet) format the display
                ConvertInput(RawDisplay.ToString)

                ' does the input fit in the size of our register
                ' v7.0.2 - 28 Jun 23: Trimming input during shortening bit size
                While cs.Stack.X.BiVal.LossOfPrecision Or Double.IsInfinity(cs.Stack.X.FVal)
                    ' remove the last digit you typed
                    RawDisplay.Remove(RawDisplay.Length - 1, 1)
                    Packet.Beep = True
                    ' try it again
                    ConvertInput(RawDisplay.ToString)
                End While

                ' Display the partial "raw" (unformatted) input.  If the input exceeds 37
                ' characters, we trim to show the right-most portion of the string.  There is
                ' no scrolling/windowing in the raw input mode. This trimming does NOT
                ' affect how the string gets converted into a number.
                If Not cs.PrgmRunning Then
                    Select Case cs.OpMode
                        Case CalcState.CalcOpMode.Float
                            Packet.DisplayText = Right(RawDisplay.ToString, 37)
                        Case CalcState.CalcOpMode.Hex
                            Packet.DisplayText = String.Format("{0} h", Right(RawDisplay.ToString, 37))
                        Case CalcState.CalcOpMode.Dec
                            Packet.DisplayText = String.Format("{0} d", Right(RawDisplay.ToString, 37))
                        Case CalcState.CalcOpMode.Oct
                            Packet.DisplayText = String.Format("{0} o", Right(RawDisplay.ToString, 37))
                        Case CalcState.CalcOpMode.Bin
                            Packet.DisplayText = String.Format("{0} b", Right(RawDisplay.ToString, 37)
                                                               )
                    End Select
                End If
            Else
                ' display the formatted value of the X register
                If Not cs.PrgmRunning Then
                    Packet.DisplayText = FormatDisplay(cs.OpMode, WinPos)
                End If
            End If

        End If

        ' update the packet info
        Packet.Carry_Annunciator = cs.Flags(CalcState.CalcFlag.Carry)
        Packet.Overflow_Annunciator = cs.Flags(CalcState.CalcFlag.Overflow)
        Return Packet
    End Function

    ''' <summary>
    ''' Parse the raw input based upon the current operating mode
    ''' </summary>
    ''' <param name="text">The Raw input</param>
    ''' <remarks>No error catching done here</remarks>
    Private Sub ConvertInput(text As String)
        Dim temp As New Register(cs.WordSize, cs.ArithMode)

        Select Case cs.OpMode
            Case CalcState.CalcOpMode.Float
                ' make an incomplete Scientific Notation parseable (is that a word?)
                If text.EndsWith("e") Then
                    text &= "0"
                End If
                Try
                    ' v6.0.7 - 22 Jul 2015
                    temp.FVal = Double.Parse(text, Globalization.NumberStyles.Float, System.Globalization.CultureInfo.InvariantCulture)
                Catch
                    ' We need some sort of marker to show that an error occurred
                    temp.FVal = Double.PositiveInfinity
                End Try
            Case CalcState.CalcOpMode.Hex
                temp.BiVal = New BigInt("&H" & text, cs.WordSize, cs.ArithMode)
            Case CalcState.CalcOpMode.Dec
                temp.BiVal = New BigInt(text, cs.WordSize, cs.ArithMode)
            Case CalcState.CalcOpMode.Oct
                temp.BiVal = New BigInt("&O" & text, cs.WordSize, cs.ArithMode)
            Case CalcState.CalcOpMode.Bin
                ' Note: the "&B" is a made up notation
                temp.BiVal = New BigInt("&B" & text, cs.WordSize, cs.ArithMode)
        End Select

        cs.Stack.X = temp
    End Sub

    ''' <summary>
    ''' Format the value of the X register for the display using
    ''' the current operating mode
    ''' </summary>
    ''' <remarks>Overloaded</remarks>
    Private Function FormatDisplay() As String
        Return FormatDisplay(cs.OpMode, 0)
    End Function

    ''' <summary>
    ''' Format the value of the X register for the display using
    ''' the specified operating mode, window, and scrolling
    ''' </summary>
    ''' <remarks>Overloaded</remarks>
    Private Function FormatDisplay(Op As CalcState.CalcOpMode, Position As Integer) As String
        Dim DisplayText, temp As String

        DisplayText = ""
        Select Case Op
            Case CalcState.CalcOpMode.Float
                ' format the screen as scientific notation?
                Dim Formatter As String
                If cs.FloatPrecision = k.KeyDp Then
                    ' You can't adjust the precision of scientific notation in
                    ' the real calculator, so I just picked 15... I hope you like it
                    Formatter = "{0:e15}"
                Else
                    Formatter = "{0:f" & cs.FloatPrecision.ToString & "}"
                End If
                DisplayText = String.Format(Formatter, cs.Stack.X.FVal)
                ' Check to see if we need to automatically switch to scientific notation
                ' Note: this is for display purposes only, the float/EEX mode does not change
                'If DisplayText.Length > 37 Or (DisplayText = "0." & StrDup(cs.FloatPrecision, "0") And cs.Stack.X.FVal <> 0.0) Then
                If DisplayText.Length > 37 Then
                    DisplayText = String.Format("{0:e15}", cs.Stack.X.FVal)
                End If
            Case CalcState.CalcOpMode.Hex
                temp = cs.Stack.X.BiVal.ToStringHex.ToUpper
                If Not cs.Flags(CalcState.CalcFlag.LeadingZero) Then
                    temp = TrimZeros(temp)
                End If
                DisplayText = String.Format("{0} h", temp)
            Case CalcState.CalcOpMode.Dec
                ' Note: the decimal mode is the only one that has to deal with the Arithmetic mode
                temp = TrimZeros(cs.Stack.X.BiVal.ToStringDec)
                ' It's rather odd, that the LeadingZero feature is forced off (page 36)
                DisplayText = String.Format("{0} d", temp)
            Case CalcState.CalcOpMode.Oct
                temp = cs.Stack.X.BiVal.ToStringOct
                If Not cs.Flags(CalcState.CalcFlag.LeadingZero) Then
                    temp = TrimZeros(temp)
                End If
                DisplayText = String.Format("{0} o", temp)
            Case CalcState.CalcOpMode.Bin
                temp = cs.Stack.X.BiVal.ToStringBin(True)
                ' Note: the binary mode is the only one that supports windows/scrolling

                If temp.Length > 35 Then
                    If Position = 0 Then
                        ' Window 0
                        ' Note: We ignore the Leading Zero flag for window 0
                        DisplayText = String.Format("{0} .b", Right(temp, 35))
                    ElseIf Position = 36 Then
                        ' Window 1
                        temp = temp.Substring(0, temp.Length - 36)
                        If Not cs.Flags(CalcState.CalcFlag.LeadingZero) Then
                            temp = TrimZeros(temp)
                        End If
                        DisplayText = String.Format("{0} b.", temp)
                    Else
                        ' TODO: Consider scrolling by whole bytes?
                        ' Note: This is not exactly the way that the real calculator does
                        ' things... it allows you to "walk off the edge" of the left-most
                        ' bits, whereas I just stop at the edge of the screen.
                        Dim pos As Integer
                        pos = (temp.Length - 35 - Position)
                        temp = temp.Substring(pos, 35)
                        ' We ignore the Leading Zero flag while scrolling
                        DisplayText = String.Format("{0} .b.", temp)
                    End If
                Else
                    If Not cs.Flags(CalcState.CalcFlag.LeadingZero) Then
                        temp = TrimZeros(temp)
                    End If
                    DisplayText = String.Format("{0} b", temp)
                End If
        End Select

        Return DisplayText
    End Function

    ''' <summary>
    ''' Locate the line number for a given program label
    ''' </summary>
    ''' <param name="start">The starting line location</param>
    ''' <param name="label">The label to search for</param>
    ''' <returns>The line number of the label</returns>
    ''' <remarks>returns a -1 if not found</remarks>
    Private Function FindLabel(start As Integer, label As Integer) As Integer
        Dim line, Instruction As String

        ' a quick sanity check...
        ' v6.0.2 - 26 Apr 12
        ' v6.0.5 - 14 Aug 13
        If (start < 0 Or start > cs.PrgmMemory.Count Or cs.PrgmMemory.Count = 0) Then
            Return -1
        End If

        ' build a prototype "g LBL" instruction and append the label
        Instruction = "43,22, " & Hex(label)

        ' search for that string from the current position
        For i As Integer = start To cs.PrgmMemory.Count - 1
            line = cs.PrgmMemory(i)
            If line.StartsWith(Instruction) Then
                Return i
            End If
        Next

        ' searches are allowed to wrap around
        For i As Integer = 0 To start - 1
            line = cs.PrgmMemory(i)
            If line.StartsWith(Instruction) Then
                Return i
            End If
        Next

        Return -1
    End Function

    ''' <summary>
    ''' Trim leading zeros from a string
    ''' </summary>
    ''' <param name="buf">The formatted display</param>
    ''' <returns>The string with leading zero's removed</returns>
    ''' <remarks>Also removed spaces</remarks>
    Private Shared Function TrimZeros(ByVal buf As String) As String
        Dim i As Integer = 0
        Dim b As String
        Dim is_neg As Boolean

        is_neg = False
        If buf.StartsWith("-") Then
            i = 1
            is_neg = True
        End If

        Do
            b = buf.Substring(i, 1)
            i = i + 1
            If i >= buf.Length Then
                Exit Do
            End If
        Loop While (b = "0" Or b = " ")
        i = i - 1

        ' Leave at least the last zero
        If i = buf.Length Then
            i = i - 1
        End If

        If is_neg Then
            Return "-" & buf.Substring(i)
        Else
            Return buf.Substring(i)
        End If
    End Function

    ''' <summary>
    ''' Provide the ability to copy text directly into the RawDisplay
    ''' </summary>
    ''' <param name="text"></param>
    ''' <remarks></remarks>
    Public Sub ImportRawDisplay(text As String)

        Try
            ConvertInput(text)
        Catch ex As Exception
            RawDisplay.Length = 0
            Throw New FormatException("Can't import pasted data", ex)
        End Try

        ' does the input fit in the size of our register
        If cs.Stack.X.BiVal.LossOfPrecision Or Double.IsInfinity(cs.Stack.X.FVal) Then
            RawDisplay.Length = 0
            Throw New ArgumentOutOfRangeException("text", "Imported data out of range")
        End If

        FormatDisplay()
    End Sub


    ' A set of methods to record the key sequences into a program line.  There are 4 versions
    ' (for whatever reason) to account for differing parameter counts

    Private Sub recordLine(key As String)
        cs.PrgmMemory.Insert(cs.PrgmPosition, "       " + key)
        cs.PrgmPosition += 1
    End Sub

    Private Sub recordLine(keycode As Integer, key As String)
        cs.PrgmMemory.Insert(cs.PrgmPosition, String.Format("{0,8:X}    '{1}", keycode, key))
        cs.PrgmPosition += 1
    End Sub

    Private Sub recordLine(prefix As Integer, keycode As Integer, key As String)
        cs.PrgmMemory.Insert(cs.PrgmPosition, String.Format("{0,5:X}{1,3:X1}    '{2}", prefix, keycode, key))
        cs.PrgmPosition += 1
    End Sub

    Private Sub recordLine(prefix As Integer, keycode As Integer, val As Integer, key As String)
        ' rather odd that they use commas here...
        cs.PrgmMemory.Insert(cs.PrgmPosition, String.Format("{0,2:X},{1,2:X},{2,2:X}    '{3}", prefix, keycode, val, key))
        cs.PrgmPosition += 1
    End Sub

End Class