﻿Imports System.Runtime.InteropServices
Imports System.IO
Imports System.Drawing.Printing

''' <summary>
''' The Main form for the application
''' </summary>
''' <remarks></remarks>
Public Class fmMain

    Const CALC_WIDTH As Integer = 512
    Const CALC_HEIGHT As Integer = 320
    Const DISPLAY_WIDTH As Integer = 320
    Const DISPLAY_HEIGHT As Integer = 38
    Const BUTTON_WIDTH As Integer = 37
    Const BUTTON_HEIGHT As Integer = 33

    Private cs As CalcState
    Private c As Calculator

    Private cached_width As Integer = 0
    Private cached_height As Integer = 0

    ''' <summary>
    ''' The form load event handler for the main form.
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub fmMain_Load(sender As System.Object, e As System.EventArgs) Handles MyBase.Load
        ' set the click event handlers for all the GButtons
        For Each ct As Control In pnCalcFace.Controls
            If TypeOf ct Is GButton Then
                AddHandler ct.Click, AddressOf GButton_Click
            End If
        Next

        ' load user's version of the saved calculator state
        cs = New CalcState
        c = New Calculator(cs)

        ' v7.0.0 - 6 Jun 21: New AutoScale feature
        Dim fontsize As Integer = CInt(My.Settings.FontSize)
        If fontsize <> 0 Then
            tbDisplay.AutoScale = False
            Dim tbFont As Font = New Font(tbDisplay.Font.FontFamily, fontsize)
            tbDisplay.Font = tbFont
        End If
        LoadState()
        UpdateMenu()

        ' Process a dummy key to refresh the display
        ProcessPacket(c.ProcessKey(-1))
    End Sub

    ''' <summary>
    ''' Handle the manual resizing of the main form
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks>Note: Autoscale Mode is set to None</remarks>
    Private Sub fmMain_SizeChanged(sender As System.Object, e As System.EventArgs) Handles MyBase.SizeChanged
        Dim x, y, width, height As Integer

        If Me.Width <> cached_width Or Me.Height <> cached_height Then
            cached_width = Me.Width
            cached_height = Me.Height

            ' Calculate the new size of the buttons
            width = CInt(pnCalcFace.Width * BUTTON_WIDTH / CALC_WIDTH)
            height = CInt(pnCalcFace.Height * BUTTON_HEIGHT / CALC_HEIGHT)

            ' Move and resize the buttons
            For Each ct As Control In pnCalcFace.Controls
                If TypeOf ct Is GButton Then

                    Dim bn As GButton = DirectCast(ct, GButton)
                    ' Get the xy grid location from the GButton's properties
                    x = CInt((pnCalcFace.Width * bn.X) / CALC_WIDTH)
                    y = CInt((pnCalcFace.Height * bn.Y) / CALC_HEIGHT)
                    bn.Location = New Point(x, y)

                    ' Special case:  The Enter key is taller
                    If bn.KeyCode = 54 Then
                        bn.Height = CInt(pnCalcFace.Height * 84.0 / CALC_HEIGHT)
                    Else
                        bn.Height = height
                    End If
                    bn.Width = width
                End If
            Next

            ' Move and resize the display textbox
            x = CInt(54 * pnCalcFace.Width / CALC_WIDTH)
            y = CInt(31 * pnCalcFace.Height / CALC_HEIGHT)
            width = CInt(pnCalcFace.Width * DISPLAY_WIDTH / CALC_WIDTH)
            height = CInt(pnCalcFace.Height * DISPLAY_HEIGHT / CALC_HEIGHT)

            tbDisplay.Location = New Point(x, y)
            tbDisplay.Width = width
            tbDisplay.Height = height

            Dim fontsize As Single = 5
            Dim widthLimitPixels As Integer = tbDisplay.Width - tbDisplay.Margin.Right - tbDisplay.Margin.Left
            Dim calcMax As Single = (tbDisplay.Height - tbDisplay.Margin.Top - tbDisplay.Margin.Bottom) / 1.5!

            If Not String.IsNullOrEmpty(Me.Text) Then
                Dim g As Graphics = Me.CreateGraphics
                Dim flags As TextFormatFlags = (TextFormatFlags.SingleLine And TextFormatFlags.TextBoxControl)
                Dim Proposed As Size = New Size(Integer.MaxValue, Integer.MaxValue)
                Dim f As Font = New Font(tbDisplay.Font.FontFamily, fontsize)
                Dim widthPixels As Integer = TextRenderer.MeasureText(g, "00000000 00000000 00000000 00000000 .b.", f, Proposed, flags).Width
                ' Increase font size to match

                While widthPixels < widthLimitPixels AndAlso fontsize <= 100 AndAlso fontsize <= calcMax
                    fontsize += 1
                    f = New Font(tbDisplay.Font.FontFamily, fontsize)
                    widthPixels = TextRenderer.MeasureText(g, Me.Text, f, Proposed, flags).Width
                End While

                ' this is the "calibration" font size to fit the size of tbDisplay
                fontsize -= 1.25!
                g.Dispose()

                If tbDisplay.AutoScale Then
                    ' just change the text to trigger DynamicText to do its thing
                    Dim previousText As String = tbDisplay.Text
                    tbDisplay.Text = ""
                    tbDisplay.Text = previousText
                Else
                    tbDisplay.Font = New Font(tbDisplay.Font.FontFamily, fontsize)
                End If

            End If

            Dim fAnnunciator As Font = New Font("Microsoft Sans Serif", (fontsize * 0.6!))

            ' recalculate the location for the labels
            y = CInt(55 * pnCalcFace.Height / CALC_HEIGHT)
            For Each ct As Control In pnCalcFace.Controls
                If TypeOf ct Is Label Then
                    x = CInt(CInt(ct.Tag) * pnCalcFace.Width / CALC_WIDTH)
                    ct.Location = New Point(x, y)
                    ct.Font = fAnnunciator
                End If
            Next
        End If
    End Sub

    ''' <summary>
    ''' The GButton click event handler
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks>We use one click event handler for all GButtons</remarks>
    Private Sub GButton_Click(sender As System.Object, e As System.EventArgs)
        Dim bn As GButton = DirectCast(sender, GButton)
        Dim p As DisplayPacket

        ' any keystroke will terminate a running program
        cs.PrgmRunning = False

        ' Send the keystrokes to the calculator engine
        p = c.ProcessKey(bn.KeyCode)

        ' show the results in the GUI
        ProcessPacket(p)

        ' should we start a program?
        If p.Start = DisplayPacket.StartType.RunProgram Then
            ' fire up a background thread to run our program
            Dim t As Threading.Thread = New Threading.Thread(AddressOf RunProgram)
            t.IsBackground = True
            t.Start()
        ElseIf p.Start = DisplayPacket.StartType.RunLine Then
            ' just run one line at a time
            RunLine()
        End If
    End Sub

    ''' <summary>
    ''' Update all of the GUI elements
    ''' </summary>
    ''' <param name="p"></param>
    ''' <remarks></remarks>
    Private Sub ProcessPacket(p As DisplayPacket)

        ' Display the Shifted Annunciators
        If p.F_Annunciator Then
            lbFkey.Text = "f"
        Else
            lbFkey.Text = String.Empty
        End If
        If p.G_Annunciator Then
            lbGkey.Text = "g"
        Else
            lbGkey.Text = String.Empty
        End If

        ' Display the System Flag Annunciator
        If p.Carry_Annunciator Then
            lbCarry.Text = "C"
        Else
            lbCarry.Text = String.Empty
        End If
        If p.Overflow_Annunciator Then
            lbOverflow.Text = "G"
        Else
            lbOverflow.Text = String.Empty
        End If

        ' Display the Program Annunciator
        If p.Prgm_Annunciator Then
            lbPrgm.Text = "PRGM"
        Else
            lbPrgm.Text = String.Empty
        End If

        ' Did the engine ask for a beep?
        If Not cs.PrgmRunning And p.Beep Then
            Beep()
        End If

        ' Process the menu changes (if any)
        If p.MenuNeedsUpdating Then
            UpdateMenu()
        End If

        ' Optionally show an alternate message for a short time, then
        ' go back to the standard display text
        If p.Delay > 0 Then
            SetDisplayText(p.AlternateText)
            Application.DoEvents()
            Threading.Thread.Sleep(p.Delay)
            SetDisplayText(p.DisplayText)
        Else
            If p.AlternateText = String.Empty Then
                If cs.OpMode = CalcState.CalcOpMode.Float Then
                    ' float mode is the only thing that is left justified
                    tbDisplay.TextAlign = HorizontalAlignment.Left
                End If
                SetDisplayText(p.DisplayText)
            Else
                SetDisplayText(p.AlternateText)
            End If
        End If
    End Sub

    ''' <summary>
    ''' Map a few keys to their corresponding buttons
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks>Requires the KeyPreview property to be set</remarks>
    Private Sub fmMain_KeyDown(sender As System.Object, e As System.Windows.Forms.KeyEventArgs) Handles MyBase.KeyDown
        ' Default focus is on the Enter key, so that if you press
        ' enter it will perform the function of the enter key rather
        ' than the button that last had focus
        ' v6.0.6 - 1 Mar 14
        BnEnt.Focus()
        Select Case e.KeyCode
            Case Keys.D0, Keys.NumPad0
                Bn0.PerformClick()
            Case Keys.D1, Keys.NumPad1
                Bn1.PerformClick()
            Case Keys.D2, Keys.NumPad2
                Bn2.PerformClick()
            Case Keys.D3, Keys.NumPad3
                Bn3.PerformClick()
            Case Keys.D4, Keys.NumPad4
                Bn4.PerformClick()
            Case Keys.D5, Keys.NumPad5
                Bn5.PerformClick()
            Case Keys.D6, Keys.NumPad6
                Bn6.PerformClick()
            Case Keys.D7, Keys.NumPad7
                Bn7.PerformClick()
            Case Keys.D8
                ' handle the non-keypad asterix
                If e.Modifiers = Keys.Shift Then
                    BnMul.PerformClick()
                Else
                    Bn8.PerformClick()
                End If
            Case Keys.NumPad8
                Bn8.PerformClick()
            Case Keys.D9, Keys.NumPad9
                Bn9.PerformClick()
            Case Keys.A
                BnA.PerformClick()
            Case Keys.B
                BnB.PerformClick()
            Case Keys.C
                ' v7.0.0 - 6 Jun 21: New feature - Thanks Anthony Howe
                If e.Modifiers = Keys.Control Then
                    tsmiEditCopy.PerformClick()
                Else
                    BnC.PerformClick()
                End If
            Case Keys.D
                BnD.PerformClick()
            Case Keys.E
                BnE.PerformClick()
            Case Keys.F
                BnF.PerformClick()
            Case Keys.V
                ' v7.0.0 - 6 Jun 21: New feature - Thanks Anthony Howe
                If e.Modifiers = Keys.Control Then
                    tsmiEditPaste.PerformClick()
                End If
            Case Keys.Add, Keys.Oemplus
                BnPls.PerformClick()
            Case Keys.Multiply
                BnMul.PerformClick()
            Case Keys.Enter
                BnEnt.PerformClick()
            Case Keys.Subtract, Keys.OemMinus
                BnMin.PerformClick()
            Case Keys.Decimal, Keys.OemPeriod
                BnDp.PerformClick()
            Case Keys.Divide, Keys.OemQuestion
                BnDiv.PerformClick()
            Case Keys.Back
                BnBSP.PerformClick()
            Case Keys.NumLock       ' Allow these keys to pass
            Case Keys.ShiftKey
            Case Keys.ControlKey
            Case Else
                Beep()              ' Beep on everything else
        End Select
    End Sub

    ''' <summary>
    ''' Save the calculator state when the program exits
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub fmMain_FormClosing(sender As System.Object, e As System.Windows.Forms.FormClosingEventArgs) Handles MyBase.FormClosing
        If cs.SaveOnExit Then
            ' stop any running application
            If cs.PrgmRunning Then
                cs.PrgmRunning = False
                Threading.Thread.Sleep(10)
            End If
            SaveState()
        End If
    End Sub

#Region "Programming"
    ''' <summary>
    ''' Run a program starting at the current line number
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub RunProgram()
        Do While cs.PrgmPosition < cs.PrgmMemory.Count
            ' execute the instructions
            If RunLine() Then
                ' Some error occurred
                Exit Do
            End If

            ' prepare to process the next line
            cs.PrgmPosition += 1

            ' stop if somebody pressed a key
            If cs.PrgmRunning = False Then
                Exit Do
            End If
        Loop

        ProcessPacket(c.ProcessKey(-1))
    End Sub

    ''' <summary>
    ''' Execute instructions at the current program line
    ''' </summary>
    ''' <returns>True on error</returns>
    ''' <remarks></remarks>
    Private Function RunLine() As Boolean
        Dim line As String
        Dim k1, k2, k3 As Integer
        Dim p As DisplayPacket

        ' A quick sanity check
        ' v6.0.2 - 26 Apr 12
        If (cs.PrgmPosition < 0 Or cs.PrgmPosition >= cs.PrgmMemory.Count) Then
            ' if you "step off the edge", then just stop
            Return True
        End If


        ' Go to the current line and process the keys found there.
        ' The line will be in 1 of 3 formats
        line = cs.PrgmMemory(cs.PrgmPosition)
        If line.StartsWith("      ") Then
            Try
                k1 = Integer.Parse(line.Substring(6, 2), Globalization.NumberStyles.HexNumber)
                p = c.ProcessKey(k1)
            Catch
                ' I'm anticipating that folks might edit the XML by hand
                ' to make minor tweaks to a program.  So, we have to be
                ' ready for a corrupted file
                MessageBox.Show("Corrupted Program Memory at line " & cs.PrgmPosition, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
                Return True
            End Try
        ElseIf line.StartsWith("   ") Then
            Try
                ' v7.0.0 - 6 Jun 21: Oops, I missed this
                ' special case for STO and RCL that optionally use the "." prefix
                If line(6) = "." Then
                    k1 = Integer.Parse(line.Substring(3, 2), Globalization.NumberStyles.HexNumber)
                    k2 = Integer.Parse(line.Substring(7, 1), Globalization.NumberStyles.HexNumber)
                    c.ProcessKey(k1)
                    c.ProcessKey(Calculator.k.KeyDp)
                    p = c.ProcessKey(k2)
                Else
                    k1 = Integer.Parse(line.Substring(3, 2), Globalization.NumberStyles.HexNumber)
                    k2 = Integer.Parse(line.Substring(6, 2), Globalization.NumberStyles.HexNumber)
                    c.ProcessKey(k1)
                    p = c.ProcessKey(k2)
                End If
            Catch
                MessageBox.Show("Corrupted Program Memory at line " & cs.PrgmPosition, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
                Return True
            End Try
        Else
            Try
                k1 = Integer.Parse(line.Substring(0, 2), Globalization.NumberStyles.HexNumber)
                k2 = Integer.Parse(line.Substring(3, 2), Globalization.NumberStyles.HexNumber)
                k3 = Integer.Parse(line.Substring(6, 2), Globalization.NumberStyles.HexNumber)
                c.ProcessKey(k1)
                c.ProcessKey(k2)
                p = c.ProcessKey(k3)
            Catch
                MessageBox.Show("Corrupted Program Memory at line " & cs.PrgmPosition, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
                Return True
            End Try

        End If

        ' We only update the display if there is a pause.  No Annunciator
        ' flags are updated while a program is running.
        If p.Delay > 0 Then
            If p.AlternateText = String.Empty Then
                SetDisplayText(p.DisplayText)
                Application.DoEvents()
                Threading.Thread.Sleep(p.Delay)
            Else
                SetDisplayText(p.AlternateText)
                Application.DoEvents()
                Threading.Thread.Sleep(p.Delay)
                SetDisplayText(p.DisplayText)
            End If
        End If

        Return False
    End Function

    Private Delegate Sub SetDisplayCallback(msg As String)

    ''' <summary>
    ''' Thread safe way to change the text in the display
    ''' </summary>
    ''' <param name="msg">The new text value</param>
    ''' <remarks></remarks>
    Public Sub SetDisplayText(msg As String)
        ' The following magic is required to make changing the text
        ' thread safe.
        If tbDisplay.InvokeRequired Then
            Dim d As New SetDisplayCallback(AddressOf SetDisplayText)
            Me.Invoke(d, New Object() {msg})
        Else
            tbDisplay.Text = msg
        End If
    End Sub
#End Region

#Region "Calculator State"
    ''' <summary>
    ''' Save the Calculator state to the default user's profile
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub SaveState()
        Dim ConfigFile As String

        ConfigFile = Path.Combine(Application.UserAppDataPath, "CalcState.xml")
        SaveState(ConfigFile)
    End Sub

    ''' <summary>
    ''' Save the Calculator state to a named file
    ''' </summary>
    ''' <param name="FileName">The name of the file to create/overwrite</param>
    ''' <remarks></remarks>
    Private Sub SaveState(FileName As String)
        Dim sw As IO.StreamWriter = Nothing

        Try
            sw = New IO.StreamWriter(FileName)
            sw.Write(cs.Serialize)
        Catch ex As Exception
            MessageBox.Show("Yikes!, Couldn't save the configuration file" & vbCr & ex.Message, "Error Saving Configuration", MessageBoxButtons.OK, MessageBoxIcon.Error)
            Finally
            If sw IsNot Nothing Then
                sw.Close()
            End If
        End Try
    End Sub

    ''' <summary>
    ''' Load the default Calculator state XML file
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub LoadState()
        Dim ConfigFile As String

        ConfigFile = Path.Combine(Application.UserAppDataPath, "CalcState.xml")

        ' if the user doesn't have a config file, then use the default
        If Not File.Exists(ConfigFile) Then
            ConfigFile = Path.Combine(Application.StartupPath, "CalcState.xml")

            ' if that doesn't exist, then just quit
            If Not File.Exists(ConfigFile) Then
                Exit Sub
            End If
        End If

        LoadState(ConfigFile)
    End Sub

    ''' <summary>
    ''' Load a saved Calculator state file
    ''' </summary>
    ''' <param name="FileName">The name of the existing file to read</param>
    ''' <remarks></remarks>
    Private Sub LoadState(FileName As String)
        Dim sr As IO.StreamReader = Nothing

        Try
            sr = New IO.StreamReader(FileName)
            cs.Deserialize(sr.ReadToEnd)
        Catch ex As Exception
            MessageBox.Show("Yikes!, Couldn't open the configuration file" & vbCr & ex.Message, "Error Opening Configuration", MessageBoxButtons.OK, MessageBoxIcon.Error)
        Finally
            If sr IsNot Nothing Then
                sr.Close()
            End If
        End Try
    End Sub
#End Region

#Region "Menu Items"
    Private printFont As Font
    Private streamToPrint As StringReader

    ''' <summary>
    ''' Load a previously saved configuration file
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiFileOpen_Click(sender As System.Object, e As System.EventArgs) Handles tsmiFileOpen.Click
        ofdMain.FileName = "CalcState.xml"
        ofdMain.Filter = "XML files (*.xml)|*.xml|All files (*.*)|*.*"
        ofdMain.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments)

        If ofdMain.ShowDialog = Windows.Forms.DialogResult.OK Then
            LoadState(ofdMain.FileName)
            UpdateMenu()
            ProcessPacket(c.ProcessKey(-1))
        End If
    End Sub

    ''' <summary>
    ''' Save the current "state" of the calculator
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiFileSave_Click(sender As System.Object, e As System.EventArgs) Handles tsmiFileSave.Click
        SaveState()
    End Sub

    ''' <summary>
    ''' Save a configuration to a named file
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiFileSaveAs_Click(sender As System.Object, e As System.EventArgs) Handles tsmiFileSaveAs.Click
        sfdMain.AddExtension = True
        sfdMain.DefaultExt = "xml"
        sfdMain.FileName = "CalcState.xml"
        sfdMain.Filter = "XML files (*.xml)|*.xml|All files (*.*)|*.*"
        sfdMain.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments)

        If sfdMain.ShowDialog = Windows.Forms.DialogResult.OK Then
            SaveState(sfdMain.FileName)
        End If
    End Sub

    ''' <summary>
    ''' Print a Program Listing
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks>Lifted right from the MSDN example</remarks>
    Private Sub tsmiFilePrint_Click(sender As System.Object, e As System.EventArgs) Handles tsmiFilePrint.Click
        pdlProgram.Document = pdProgram

        ' a quick sanity check...
        If cs.PrgmMemory.Count = 0 Then
            MessageBox.Show("There are no Program lines to print", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning)
            Exit Sub
        End If

        ' put a "new line" between each line
        Dim sb As New System.Text.StringBuilder
        Dim lc As Integer = 1
        For Each line As String In cs.PrgmMemory
            sb.Append(lc.ToString("000"))
            sb.Append("- ")
            sb.Append(line)
            sb.Append(vbCr)
            lc += 1
        Next

        Try
            If pdlProgram.ShowDialog() = Windows.Forms.DialogResult.OK Then
                streamToPrint = New StringReader(sb.ToString)
                printFont = New Font("Courier New", 10)
                pdProgram.Print()
            End If
        Catch ex As Exception
            MessageBox.Show("Could not print the Program" & vbCr & ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error)
        Finally
            If streamToPrint IsNot Nothing Then
                streamToPrint.Close()
            End If
        End Try

    End Sub

    ''' <summary>
    ''' The PrintPage event is raised for each page to be printed
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="ev"></param>
    ''' <remarks>Lifted right from the MSDN example</remarks>
    Private Sub pd_PrintPage(ByVal sender As Object, ByVal ev As PrintPageEventArgs) Handles pdProgram.PrintPage
        Dim linesPerPage As Single
        Dim yPos As Single
        Dim count As Integer = 0
        Dim leftMargin As Single = ev.MarginBounds.Left
        Dim topMargin As Single = ev.MarginBounds.Top
        Dim line As String = Nothing
        Dim sf As New StringFormat

        ' Calculate the number of lines per page.
        linesPerPage = ev.MarginBounds.Height / printFont.GetHeight(ev.Graphics)

        ' Print each line of the file.
        While count < linesPerPage
            line = streamToPrint.ReadLine()
            If line Is Nothing Then
                Exit While
            End If
            yPos = topMargin + count * printFont.GetHeight(ev.Graphics)
            ev.Graphics.DrawString(line, printFont, Brushes.Black, leftMargin, yPos, sf)
            count += 1
        End While
        sf.Dispose()

        ' If more lines exist, print another page.
        If (line IsNot Nothing) Then
            ev.HasMorePages = True
        Else
            ev.HasMorePages = False
        End If
    End Sub


    ''' <summary>
    ''' Printer Page Setup Dialog
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiFileSetup_Click(sender As System.Object, e As System.EventArgs) Handles tsmiFileSetup.Click
        psdProgram.Document = pdProgram
        psdProgram.ShowDialog()
    End Sub

    ''' <summary>
    ''' Let's blow this Pop stand!
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiFileExit_Click(sender As System.Object, e As System.EventArgs) Handles tsmiFileExit.Click
        Me.Close()
    End Sub

    ''' <summary>
    ''' Copy from calculator display to the clipboard
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiEditCopy_Click(sender As System.Object, e As System.EventArgs) Handles tsmiEditCopy.Click
        My.Computer.Clipboard.Clear()
        ' Hummm... should I strip the "mode indicators" off the
        ' end to make it a bit more normal?
        My.Computer.Clipboard.SetText(tbDisplay.Text.Trim)
    End Sub

    ''' <summary>
    ''' Paste from the clipboard into the display
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiEditPaste_Click(sender As System.Object, e As System.EventArgs) Handles tsmiEditPaste.Click
        If My.Computer.Clipboard.ContainsText Then
            Dim buf As String = My.Computer.Clipboard.GetText
            ' limit to 35 characters
            buf = Microsoft.VisualBasic.Right(buf, 35)
            Try
                c.ImportRawDisplay(buf)
                ' v6.0.6 - 1 Mar 14
                ' Process a dummy key to refresh the display
                ProcessPacket(c.ProcessKey(-1))
            Catch ex As Exception
                MessageBox.Show("Error Converting Pasted Data" & vbCr & ex.Message, "Paste Error", MessageBoxButtons.OK, MessageBoxIcon.Hand)
            End Try
        End If
    End Sub

    ''' <summary>
    ''' Switch to floating point mode
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiModeFloat_Click(sender As System.Object, e As System.EventArgs) Handles tsmiModeFloat.Click
        cs.OpMode = CalcState.CalcOpMode.Float
        tsmiModeFloat.Checked = True
        tsmiModeHex.Checked = False
        tsmiModeDec.Checked = False
        tsmiModeOct.Checked = False
        tsmiModeBin.Checked = False
        tsmiModeSI.Checked = False
        ProcessPacket(c.ProcessKey(-1))
    End Sub

    ''' <summary>
    ''' Switch to Hexadecimal mode
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiModeHex_Click(sender As System.Object, e As System.EventArgs) Handles tsmiModeHex.Click
        cs.OpMode = CalcState.CalcOpMode.Hex
        tsmiModeFloat.Checked = False
        tsmiModeHex.Checked = True
        tsmiModeDec.Checked = False
        tsmiModeOct.Checked = False
        tsmiModeBin.Checked = False
        tsmiModeSI.Checked = False
        ProcessPacket(c.ProcessKey(-1))
    End Sub

    ''' <summary>
    ''' Switch to Decimal mode
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiModeDec_Click(sender As System.Object, e As System.EventArgs) Handles tsmiModeDec.Click
        cs.OpMode = CalcState.CalcOpMode.Dec
        tsmiModeFloat.Checked = False
        tsmiModeHex.Checked = False
        tsmiModeDec.Checked = True
        tsmiModeOct.Checked = False
        tsmiModeBin.Checked = False
        tsmiModeSI.Checked = False
        ProcessPacket(c.ProcessKey(-1))
    End Sub

    ''' <summary>
    ''' Switch to Octal mode
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiModeOct_Click(sender As System.Object, e As System.EventArgs) Handles tsmiModeOct.Click
        cs.OpMode = CalcState.CalcOpMode.Oct
        tsmiModeFloat.Checked = False
        tsmiModeHex.Checked = False
        tsmiModeDec.Checked = False
        tsmiModeOct.Checked = True
        tsmiModeBin.Checked = False
        tsmiModeSI.Checked = False
        ProcessPacket(c.ProcessKey(-1))
    End Sub

    ''' <summary>
    ''' Switch to Binary mode
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiModeBin_Click(sender As System.Object, e As System.EventArgs) Handles tsmiModeBin.Click
        cs.OpMode = CalcState.CalcOpMode.Bin
        tsmiModeFloat.Checked = False
        tsmiModeHex.Checked = False
        tsmiModeDec.Checked = False
        tsmiModeOct.Checked = False
        tsmiModeBin.Checked = True
        tsmiModeSI.Checked = False
        ProcessPacket(c.ProcessKey(-1))
    End Sub

    ''' <summary>
    ''' Switch to Scientific Notation (float mode)
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiModeSI_Click(sender As System.Object, e As System.EventArgs) Handles tsmiModeSI.Click
        cs.OpMode = CalcState.CalcOpMode.Float
        tsmiModeFloat.Checked = True
        tsmiModeHex.Checked = False
        tsmiModeDec.Checked = False
        tsmiModeOct.Checked = False
        tsmiModeBin.Checked = False
        If cs.FloatPrecision = Calculator.k.KeyDp Then
            cs.FloatPrecision = CInt(tscbModePrecision.SelectedItem)
            tsmiModeSI.Checked = False
        Else
            cs.FloatPrecision = Calculator.k.KeyDp
            tsmiModeSI.Checked = True
        End If
        ProcessPacket(c.ProcessKey(-1))
    End Sub

    ''' <summary>
    ''' Set the number of decimal points in float mode
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tscbModePrecision_SelectedIndexChanged(sender As System.Object, e As System.EventArgs) Handles tscbModePrecision.SelectedIndexChanged
        cs.FloatPrecision = CInt(tscbModePrecision.SelectedItem)
        tsmiModeFloat.Checked = True
        tsmiModeHex.Checked = False
        tsmiModeDec.Checked = False
        tsmiModeOct.Checked = False
        tsmiModeBin.Checked = False
        tsmiModeSI.Checked = False
        tsmiMode.HideDropDown()
        ProcessPacket(c.ProcessKey(-1))
    End Sub

    ''' <summary>
    ''' Reset the configuration to the "factory defaults"
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiOptionsClear_Click(sender As System.Object, e As System.EventArgs) Handles tsmiOptionsClear.Click
        Dim ConfigFile As String = IO.Path.Combine(Application.StartupPath, "CalcState.xml")
        LoadState(ConfigFile)
        UpdateMenu()
        ProcessPacket(c.ProcessKey(-1))
    End Sub

    ''' <summary>
    ''' Should we automatically save the configuration upon exit?
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiOptionsSave_Click(sender As System.Object, e As System.EventArgs) Handles tsmiOptionsSave.Click
        cs.SaveOnExit = Not cs.SaveOnExit
        tsmiOptionsSave.Checked = cs.SaveOnExit
    End Sub

    ''' <summary>
    ''' Set word size to 8 bit
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiOptions8bit_Click(sender As System.Object, e As System.EventArgs) Handles tsmiOptions8bit.Click
        cs.WordSize = 8
        tsmiOptions8bit.Checked = True
        tsmiOptions16bit.Checked = False
        tsmiOptions32bit.Checked = False
        tsmiOptions64bit.Checked = False
        ProcessPacket(c.ProcessKey(-1))
    End Sub

    ''' <summary>
    ''' Set word size to 16 bit
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiOptions16bit_Click(sender As System.Object, e As System.EventArgs) Handles tsmiOptions16bit.Click
        cs.WordSize = 16
        tsmiOptions8bit.Checked = False
        tsmiOptions16bit.Checked = True
        tsmiOptions32bit.Checked = False
        tsmiOptions64bit.Checked = False
        ProcessPacket(c.ProcessKey(-1))
    End Sub

    ''' <summary>
    ''' Set word size to 32 bit
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiOptions32bit_Click(sender As System.Object, e As System.EventArgs) Handles tsmiOptions32bit.Click
        cs.WordSize = 32
        tsmiOptions8bit.Checked = False
        tsmiOptions16bit.Checked = False
        tsmiOptions32bit.Checked = True
        tsmiOptions64bit.Checked = False
        ProcessPacket(c.ProcessKey(-1))
    End Sub

    ''' <summary>
    ''' Set word size to 64 bit
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiOptions64bit_Click(sender As System.Object, e As System.EventArgs) Handles tsmiOptions64bit.Click
        cs.WordSize = 64
        tsmiOptions8bit.Checked = False
        tsmiOptions16bit.Checked = False
        tsmiOptions32bit.Checked = False
        tsmiOptions64bit.Checked = True
        ProcessPacket(c.ProcessKey(-1))
    End Sub

    ''' <summary>
    ''' Set Arithmetic (complement) mode to 1's complement
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiOptions1sComp_Click(sender As System.Object, e As System.EventArgs) Handles tsmiOptions1sComp.Click
        cs.ArithMode = BigInt.ArithMode.OnesComp
        tsmiOptions1sComp.Checked = True
        tsmiOptions2sComp.Checked = False
        tsmiOptionsUnsigned.Checked = False
        ProcessPacket(c.ProcessKey(-1))
    End Sub

    ''' <summary>
    ''' Set Arithmetic (complement) mode to 2's complement
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiOptions2sComp_Click(sender As System.Object, e As System.EventArgs) Handles tsmiOptions2sComp.Click
        cs.ArithMode = BigInt.ArithMode.TwosComp
        tsmiOptions1sComp.Checked = False
        tsmiOptions2sComp.Checked = True
        tsmiOptionsUnsigned.Checked = False
        ProcessPacket(c.ProcessKey(-1))
    End Sub

    ''' <summary>
    ''' Set Arithmetic (complement) mode to unsigned
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiOptionsUnsigned_Click(sender As System.Object, e As System.EventArgs) Handles tsmiOptionsUnsigned.Click
        cs.ArithMode = BigInt.ArithMode.Unsigned
        tsmiOptions1sComp.Checked = False
        tsmiOptions2sComp.Checked = False
        tsmiOptionsUnsigned.Checked = True
        ProcessPacket(c.ProcessKey(-1))
    End Sub

    ''' <summary>
    ''' Set/Clear User flag 0
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiFlagsUser0_Click(sender As System.Object, e As System.EventArgs) Handles tsmiFlagsUser0.Click
        cs.Flags(CalcState.CalcFlag.User0) = Not cs.Flags(CalcState.CalcFlag.User0)
        tsmiFlagsUser0.Checked = cs.Flags(CalcState.CalcFlag.User0)
    End Sub

    ''' <summary>
    ''' Set/Clear User flag 1
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiFlagsUser1_Click(sender As System.Object, e As System.EventArgs) Handles tsmiFlagsUser1.Click
        cs.Flags(CalcState.CalcFlag.User1) = Not cs.Flags(CalcState.CalcFlag.User1)
        tsmiFlagsUser1.Checked = cs.Flags(CalcState.CalcFlag.User1)
    End Sub

    ''' <summary>
    ''' Set/Clear User flag 2
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiFlagsUser2_Click(sender As System.Object, e As System.EventArgs) Handles tsmiFlagsUser2.Click
        cs.Flags(CalcState.CalcFlag.User2) = Not cs.Flags(CalcState.CalcFlag.User2)
        tsmiFlagsUser2.Checked = cs.Flags(CalcState.CalcFlag.User2)
    End Sub

    ''' <summary>
    ''' Set/Clear System Flag 3 (display leading zeros)
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiFlagsZeros_Click(sender As System.Object, e As System.EventArgs) Handles tsmiFlagsZeros.Click
        cs.Flags(CalcState.CalcFlag.LeadingZero) = Not cs.Flags(CalcState.CalcFlag.LeadingZero)
        tsmiFlagsZeros.Checked = cs.Flags(CalcState.CalcFlag.LeadingZero)
        ProcessPacket(c.ProcessKey(-1))
    End Sub

    ''' <summary>
    ''' Set/Clear System Flag 4 (carry bit)
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiFlagsCarry_Click(sender As System.Object, e As System.EventArgs) Handles tsmiFlagsCarry.Click
        cs.Flags(CalcState.CalcFlag.Carry) = Not cs.Flags(CalcState.CalcFlag.Carry)
        tsmiFlagsCarry.Checked = cs.Flags(CalcState.CalcFlag.Carry)
        ProcessPacket(c.ProcessKey(-1))
    End Sub

    ''' <summary>
    ''' Set/Clear System Flag 5 (overflow)
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiFlagsOverflow_Click(sender As System.Object, e As System.EventArgs) Handles tsmiFlagsOverflow.Click
        cs.Flags(CalcState.CalcFlag.Overflow) = Not cs.Flags(CalcState.CalcFlag.Overflow)
        tsmiFlagsOverflow.Checked = cs.Flags(CalcState.CalcFlag.Overflow)
        ProcessPacket(c.ProcessKey(-1))
    End Sub

    ''' <summary>
    ''' Display Help table of contents
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiHelpContents_Click(sender As System.Object, e As System.EventArgs) Handles tsmiHelpContents.Click
        Help.ShowHelp(Me, "wrpn.chm")
    End Sub

    ''' <summary>
    ''' Display Help index
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiHelpIndex_Click(sender As System.Object, e As System.EventArgs) Handles tsmiHelpIndex.Click
        Help.ShowHelpIndex(Me, "wrpn.chm")
    End Sub

    ''' <summary>
    ''' Display an image of the back panel
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiHelpBackPanel_Click(sender As System.Object, e As System.EventArgs) Handles tsmiHelpBackPanel.Click
        Dim BackPanel As New fmBackPanel

        BackPanel.ShowDialog()
    End Sub

    ''' <summary>
    ''' Display the About box
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tsmiHelpAbout_Click(sender As System.Object, e As System.EventArgs) Handles tsmiHelpAbout.Click
        Dim About As New fmAbout

        About.ShowDialog()
    End Sub

    ''' <summary>
    ''' Update all of the checkbox values in the menu
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub UpdateMenu()

        ' The View menu
        tsmiModeFloat.Checked = False
        tsmiModeHex.Checked = False
        tsmiModeDec.Checked = False
        tsmiModeOct.Checked = False
        tsmiModeBin.Checked = False
        tsmiModeSI.Checked = False
        Select Case cs.OpMode
            Case CalcState.CalcOpMode.Float
                tsmiModeFloat.Checked = True
                If cs.FloatPrecision = Calculator.k.KeyDp Then
                    tsmiModeSI.Checked = True
                Else
                    tsmiModeSI.Checked = False
                End If
            Case CalcState.CalcOpMode.Hex
                tsmiModeHex.Checked = True
            Case CalcState.CalcOpMode.Dec
                tsmiModeDec.Checked = True
            Case CalcState.CalcOpMode.Oct
                tsmiModeOct.Checked = True
            Case CalcState.CalcOpMode.Bin
                tsmiModeBin.Checked = True
        End Select
        If cs.FloatPrecision <> Calculator.k.KeyDp Then
            tscbModePrecision.Text = cs.FloatPrecision.ToString
        End If

        ' The Options Menu
        tsmiOptionsSave.Checked = cs.SaveOnExit
        tsmiOptions8bit.Checked = False
        tsmiOptions16bit.Checked = False
        tsmiOptions32bit.Checked = False
        tsmiOptions64bit.Checked = False
        Select Case cs.WordSize
            Case 8
                tsmiOptions8bit.Checked = True
            Case 16
                tsmiOptions16bit.Checked = True
            Case 32
                tsmiOptions32bit.Checked = True
            Case 64
                tsmiOptions64bit.Checked = True
            Case Else
                ' unlike a lot of other menu items, this one may
                ' not have anything checked
        End Select
        tsmiOptions1sComp.Checked = False
        tsmiOptions2sComp.Checked = False
        tsmiOptionsUnsigned.Checked = False
        Select Case cs.ArithMode
            Case BigInt.ArithMode.OnesComp
                tsmiOptions1sComp.Checked = True
            Case BigInt.ArithMode.TwosComp
                tsmiOptions2sComp.Checked = True
            Case BigInt.ArithMode.Unsigned
                tsmiOptionsUnsigned.Checked = True
        End Select

        ' The Flags Menu
        tsmiFlagsUser0.Checked = cs.Flags(CalcState.CalcFlag.User0)
        tsmiFlagsUser1.Checked = cs.Flags(CalcState.CalcFlag.User1)
        tsmiFlagsUser2.Checked = cs.Flags(CalcState.CalcFlag.User2)
        tsmiFlagsZeros.Checked = cs.Flags(CalcState.CalcFlag.LeadingZero)
        tsmiFlagsCarry.Checked = cs.Flags(CalcState.CalcFlag.Carry)
        tsmiFlagsOverflow.Checked = cs.Flags(CalcState.CalcFlag.Overflow)
    End Sub
#End Region

    ' This will make the resizing Of the calculator a lot less resource intensive
    ' https://stackoverflow.com/questions/29548479/turn-off-show-window-contents-while-dragging-setting-while-displaying-my-c-sha

    <DllImport("user32.dll", EntryPoint:="SystemParametersInfo", CharSet:=CharSet.Auto)>
    Public Shared Function GetSystemParametersInfo(ByVal uAction As Integer, ByVal uParam As Integer, ByRef lpvParam As Integer, ByVal fuWinIni As Integer) As Integer
    End Function

    <DllImport("user32.dll", EntryPoint:="SystemParametersInfo", CharSet:=CharSet.Auto)>
    Public Shared Function SetSystemParametersInfo(ByVal uAction As Integer, ByVal uParam As Integer, ByVal lpvParam As Integer, ByVal fuWinIni As Integer) As Integer
    End Function

    Private Const SPI_GETDRAGFULLWINDOWS As Integer = 38
    Private Const SPI_SETDRAGFULLWINDOWS As Integer = 37
    Private Const WM_SYSCOMMAND As Integer = 274
    Private Const SC_SIZE As Integer = 61440

    'change 'Show window content while dragging' to false while resizing
    Protected Overrides Sub WndProc(ByRef m As Message)
        If m.Msg = WM_SYSCOMMAND AndAlso (m.WParam.ToInt32 And 65520) = SC_SIZE Then
            Dim isDragFullWindow As Integer

            GetSystemParametersInfo(SPI_GETDRAGFULLWINDOWS, 0, isDragFullWindow, 0)
            If isDragFullWindow <> 0 Then
                SetSystemParametersInfo(SPI_SETDRAGFULLWINDOWS, 0, 0, 0)
            End If

            MyBase.WndProc(m)
            If isDragFullWindow <> 0 Then
                SetSystemParametersInfo(SPI_SETDRAGFULLWINDOWS, 1, 0, 0)
            End If

        Else
            MyBase.WndProc(m)
        End If

    End Sub

    'reduce control flickering and black stripes when window is resized
    Protected Overrides ReadOnly Property CreateParams As CreateParams
        Get
            Dim cp As CreateParams = MyBase.CreateParams
            cp.ExStyle = (cp.ExStyle Or 33554432)
            ' Turn on WS_EX_COMPOSITED
            Return cp
        End Get
    End Property

End Class
